# ======================================================================
# Copyright 2016 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import time
from typing import Type

from supervisor.compat import as_bytes, as_string
from supervisor.states import RUNNING_STATES, STOPPED_STATES
from supervisor.web import MeldView

from supvisors import __version__
from supvisors.instancestatus import SupvisorsInstanceStatus
from supvisors.internal_com.mapper import SupvisorsInstanceId
from supvisors.statscompiler import ProcStatisticsInstance
from supvisors.ttypes import SupvisorsStates, Payload, PayloadList
from supvisors.utils import get_stats, get_small_value
from .viewcontext import *
from .viewimage import process_cpu_img, process_mem_img, SoftwareIconImage
from .webutils import *


class ViewHandler(MeldView):
    """ Helper class to share rendering and behavior between handlers inheriting from MeldView. """

    def __init__(self, context):
        """ Initialization of the attributes. """
        MeldView.__init__(self, context)
        self.page_name = None
        self.current_mtime = time.monotonic()
        self.current_time = time.time()
        # add Supvisors shortcuts
        self.supvisors = context.supervisord.supvisors
        self.logger = self.supvisors.logger
        # cannot store context as it is named, or it would crush the http context
        self.sup_ctx = self.supvisors.context
        # even if there is no local collector, statistics can be available from other Supvisors instances
        # where a collector is available
        self.has_host_statistics = True
        self.has_process_statistics = True
        # init view_ctx (only for tests)
        self.view_ctx: Optional[ViewContext] = None

    @property
    def local_identifier(self):
        """ Return the identifier of the local Supvisors instance. """
        return self.supvisors.mapper.local_identifier

    @property
    def local_nick_identifier(self):
        """ Return the nick identifier of the local Supvisors instance. """
        return self.supvisors.mapper.local_nick_identifier

    def __call__(self):
        """ Anticipation of Supervisor#1273.
        Return response body as bytes instead of as string to prevent UnicodeDecodeError in the event
        of using binary references (images) in HTML. """
        response = MeldView.__call__(self)
        if response is NOT_DONE_YET:
            return NOT_DONE_YET
        response['body'] = as_bytes(response['body'])
        return response

    def render(self):
        """ Handle the rendering of the Supvisors pages. """
        # manage parameters
        self.handle_parameters()
        # manage action
        message = self.handle_action()
        if message is NOT_DONE_YET:
            return NOT_DONE_YET
        # clone the template and set navigation menu
        root = self.clone()
        # write navigation menu, page header and contents
        self.write_style(root)
        self.write_common(root)
        self.write_navigation(root)
        # write the header section
        header_elt = root.findmeld('header_mid')
        self.write_header(header_elt)
        # write the body section
        contents_elt = root.findmeld('contents_mid')
        self.write_contents(contents_elt)
        # send message only at the end to get all URL parameters
        self.view_ctx.fire_message()
        return as_string(root.write_xhtmlstring())

    def handle_parameters(self):
        """ Retrieve the parameters selected on the web page. """
        self.view_ctx = ViewContext(self.context)
        self.logger.debug(f'ViewHandler.handle_parameters: new context {self.view_ctx.parameters}')

    def write_style(self, root):
        """ Entry point for additional style instructions. """

    def write_common(self, root):
        """ Common rendering of the Supvisors pages. """
        # set auto-refresh status on page
        auto_refresh = self.view_ctx.auto_refresh
        if not auto_refresh:
            root.findmeld('meta_mid').deparent()
        # blink main title in conciliation state
        if self.supvisors.fsm.state == SupvisorsStates.CONCILIATION:
            elt = root.findmeld('supvisors_mid')
            update_attrib(elt, 'class', 'failure')
        # set Supvisors version
        root.findmeld('version_mid').content(__version__)
        # set bottom message
        footer_elt = root.findmeld('footer_mid')
        print_message(footer_elt, self.view_ctx.gravity, self.view_ctx.message,
                      self.current_time, self.local_nick_identifier)

    def write_navigation(self, root):
        """ Write the navigation menu.
        Subclasses will define the write_nav parameters to be used. """
        raise NotImplementedError

    def write_nav(self, root, identifier: Optional[str] = None, appli: Optional[str] = None,
                  source: Optional[str] = None):
        """ Write the navigation menu. """
        self.write_nav_instances(root, identifier, source)
        self.write_nav_applications(root, appli)

    def write_nav_instances(self, root, identifier: Optional[str], source: Optional[str]) -> None:
        """ Write the node part of the navigation menu. """
        mid_elt = root.findmeld('instance_li_mid')
        identifiers = list(self.supvisors.mapper.instances.keys())
        any_failure = False
        # in discovery mode, other Supvisors instances arrive randomly in every Supvisors instance
        # so let's sort them by name
        if self.supvisors.options.discovery_mode:
            identifiers = [status.identifier
                           for status in sorted(self.supvisors.mapper.instances.values(),
                                                key=lambda x: x.nick_identifier)]
        for li_elt, item in mid_elt.repeat(identifiers):
            try:
                status: SupvisorsInstanceStatus = self.sup_ctx.instances[item]
            except KeyError:
                self.logger.debug(f'ViewHandler.write_nav_instances: failed to get instance status from {item}')
            else:
                failure = status.has_error()
                any_failure |= failure
                # set element class
                update_attrib(li_elt, 'class', status.state.name)
                if item == identifier:
                    update_attrib(li_elt, 'class', 'active')
                elif item == source:
                    update_attrib(li_elt, 'class', 'local')
                if failure:
                    update_attrib(li_elt, 'class', 'failure')
                # set hyperlink attributes
                elt = li_elt.findmeld('instance_a_mid')
                if status.state_modes.starting_jobs or status.state_modes.stopping_jobs:
                    update_attrib(elt, 'class', 'blink')
                if status.has_active_state():
                    # go to web page located on the Supvisors instance to reuse Supervisor StatusView
                    url = self.view_ctx.format_url(item, PROC_INSTANCE_PAGE)
                    elt.attributes(href=url)
                    update_attrib(elt, 'class', 'on')
                else:
                    update_attrib(elt, 'class', 'off')
                # set content (master and failure symbols need a positional update against the text)
                instance_elt = elt.findmeld('instance_sp_mid')
                instance_elt.content(status.supvisors_id.nick_identifier)
                if item == self.sup_ctx.master_identifier:
                    master_elt = elt.findmeld('master_sp_mid')
                    master_elt.content(MASTER_SYMBOL)
                    update_attrib(li_elt, 'class', 'master')
        # warn at title level if any application has a failure
        if any_failure:
            elt = root.findmeld('instance_h_mid')
            update_attrib(elt, 'class', 'failure')

    def write_nav_applications(self, root, appli):
        """ Write the application part of the navigation menu. """
        any_failure = False
        # write applications
        mid_elt = root.findmeld('appli_li_mid')
        applications = self.sup_ctx.get_managed_applications().values()
        working_apps = (self.supvisors.starter.get_application_job_names()
                        | self.supvisors.stopper.get_application_job_names())
        # forced to list otherwise not easily testable
        for li_elt, item in mid_elt.repeat(sorted(applications, key=lambda x: x.application_name)):
            failure = item.major_failure or item.minor_failure
            any_failure |= failure
            # set element class
            update_attrib(li_elt, 'class', item.state.name)
            if item.application_name == appli:
                update_attrib(li_elt, 'class', 'active')
            if failure:
                update_attrib(li_elt, 'class', 'failure')
            # set hyperlink attributes
            elt = li_elt.findmeld('appli_a_mid')
            if item.application_name in working_apps:
                update_attrib(elt, 'class', 'blink')
            if self.supvisors.fsm.state in [SupvisorsStates.OFF, SupvisorsStates.INITIALIZATION]:
                update_attrib(elt, 'class', 'off')
            else:
                # force default application starting strategy
                url = self.view_ctx.format_url('', APPLICATION_PAGE, **{APPLI: item.application_name,
                                                                        STRATEGY: item.rules.starting_strategy.name})
                elt.attributes(href=url)
                update_attrib(elt, 'class', 'on')
            elt.content(item.application_name)
        # warn at title level if any application has a failure
        if any_failure:
            elt = root.findmeld('appli_h_mid')
            update_attrib(elt, 'class', 'failure')

    def write_header(self, header_elt):
        """ Write the header common part of the page.
        Subclasses will define what's to be done at their level. """
        self.write_software(header_elt)
        self.write_status(header_elt)
        self.write_options(header_elt)
        self.write_actions(header_elt)

    def write_software(self, header_elt):
        """ Write the user software elements. """
        card_elt = header_elt.findmeld('software_card_mid')
        if not (self.supvisors.options.software_name or self.supvisors.options.software_icon):
            # remove the user software card if nothing set
            card_elt.replace('')
        else:
            if self.supvisors.options.software_name:
                # write software name
                card_elt.findmeld('software_name_mid').content(self.supvisors.options.software_name)
            # write user icon
            if self.supvisors.options.software_icon:
                SoftwareIconImage.set_path(self.supvisors.options.software_icon)
            else:
                card_elt.findmeld('software_icon_mid').replace('')

    def write_status(self, header_elt):
        """ Write the page instance status. """
        raise NotImplementedError

    def write_options(self, header_elt):
        """ Write configured periods for statistics.
        Does nothing by default. To be specialized in subclasses where statistics are available. """

    def write_periods(self, header_elt) -> None:
        """ Write configured periods for statistics.
        The update is conditioned to statistics options set in subclasses. """
        # write the available periods
        mid_elt = header_elt.findmeld('period_li_mid')
        for li_elt, item in mid_elt.repeat(self.supvisors.options.stats_periods):
            # print period button
            elt = li_elt.findmeld('period_a_mid')
            if item == self.view_ctx.period:
                update_attrib(li_elt, 'class', 'off active')
            else:
                update_attrib(li_elt, 'class', 'on')
                url = self.view_ctx.format_url('', self.page_name, **{PERIOD: item})
                elt.attributes(href=url)
            elt.content(f'{item}s')

    def write_actions(self, header_elt):
        """ Write the common action elements. """
        # configure refresh button
        elt = header_elt.findmeld('refresh_a_mid')
        url = self.view_ctx.format_url('', self.page_name)
        elt.attributes(href=url)
        # configure auto-refresh button
        auto_refresh = self.view_ctx.auto_refresh
        elt = header_elt.findmeld('autorefresh_a_mid')
        url = self.view_ctx.format_url('', self.page_name, **{AUTO: not auto_refresh})
        elt.attributes(href=url)
        if auto_refresh:
            update_attrib(elt, 'class', 'active')

    def write_contents(self, root):
        """ Write the contents part of the page.
        Subclasses will define what's to be done. """
        raise NotImplementedError

    def write_global_shex(self, table_elt, param: str, shex: str,
                          expand_shex: bytearray, shrink_shex: bytearray) -> None:
        """ Write global shrink / expand buttons.

        :param table_elt: the table element.
        :param param: the URL attribute for shex action.
        :param shex: the current shex from context form.
        :param expand_shex: the shex key to expand all items.
        :param shrink_shex: the shex key to shrink all items.
        :return: None.
        """
        # write expand button
        elt = table_elt.findmeld('expand_a_mid')
        if shex == expand_shex.hex():
            elt.replace('')
        else:
            elt.content(SHEX_EXPAND)
            url = self.view_ctx.format_url('', self.page_name, **{param: expand_shex.hex()})
            elt.attributes(href=url)
        # write shrink button
        elt = table_elt.findmeld('shrink_a_mid')
        if shex == shrink_shex.hex():
            elt.replace('')
        else:
            elt.content(SHEX_SHRINK)
            url = self.view_ctx.format_url('', self.page_name, **{param: shrink_shex.hex()})
            elt.attributes(href=url)

    def write_common_process_cpu(self, tr_elt, info):
        """ Write the CPU part of the common process status.
        Statistics data comes from the node. """
        if self.has_process_statistics:
            proc_stats: ProcStatisticsInstance = info['proc_stats']
            elt = tr_elt.findmeld('pcpu_a_mid')
            if proc_stats and len(proc_stats.cpu) > 0:
                # print last CPU value of process
                cpuvalue = proc_stats.cpu[-1]
                if info['namespec']:  # empty for an application info
                    update_attrib(elt, 'class', 'button on')
                    parameters = {PROCESS: info['namespec'], IDENTIFIER: info['identifier']}
                    if (self.view_ctx.process_name == info['namespec']
                            and self.view_ctx.identifier == info['identifier']):
                        update_attrib(elt, 'class', 'active')
                        parameters[PROCESS] = None
                    url = self.view_ctx.format_url('', self.page_name, **parameters)
                    elt.attributes(href=url)
                    elt.content(get_small_value(cpuvalue))
                else:
                    # print data with no link and button format
                    elt.replace(get_small_value(cpuvalue))
            else:
                # when no data, do not write link
                elt.replace('--')
        else:
            # remove cell
            tr_elt.findmeld('pcpu_td_mid').deparent()

    def write_common_process_mem(self, tr_elt, info: Payload) -> None:
        """ Write the MEM part of the common process status.
        Statistics data comes from node. """
        if self.has_process_statistics:
            proc_stats: ProcStatisticsInstance = info['proc_stats']
            elt = tr_elt.findmeld('pmem_a_mid')
            if proc_stats and len(proc_stats.mem) > 0:
                # print last MEM value of process
                memvalue = proc_stats.mem[-1]
                if info['namespec']:  # empty for an application info
                    update_attrib(elt, 'class', 'button on')
                    parameters = {PROCESS: info['namespec'], IDENTIFIER: info['identifier']}
                    if (self.view_ctx.process_name == info['namespec']
                            and self.view_ctx.identifier == info['identifier']):
                        update_attrib(elt, 'class', 'active')
                        parameters[PROCESS] = None
                    url = self.view_ctx.format_url('', self.page_name, **parameters)
                    elt.attributes(href=url)
                    elt.content(get_small_value(memvalue))
                else:
                    # print data with no link
                    elt.replace(get_small_value(memvalue))
            else:
                # when no data, no not write link
                elt.replace('--')
        else:
            # remove cell
            tr_elt.findmeld('pmem_td_mid').deparent()

    def write_process_start_button(self, tr_elt, info):
        """ Write the configuration of the start button of a process.
        The action will be handled by the local Supvisors instance. """
        self._write_process_button(tr_elt, 'start_a_mid', '', self.page_name,
                                   'start', info['namespec'],
                                   info['startable'] and info['statecode'] in STOPPED_STATES)

    def write_process_stop_button(self, tr_elt, info):
        """ Write the configuration of the stop button of a process.
        The action will be handled by the local supvisors. """
        self._write_process_button(tr_elt, 'stop_a_mid', '', self.page_name,
                                   'stop', info['namespec'],
                                   info['stoppable'] and info['statecode'] in RUNNING_STATES)

    def write_process_restart_button(self, tr_elt, info):
        """ Write the configuration of the restart button of a process.
        The action will be handled by the local supvisors. """
        self._write_process_button(tr_elt, 'restart_a_mid', '', self.page_name,
                                   'restart', info['namespec'],
                                   info['startable'] and info['statecode'] in RUNNING_STATES)

    def write_process_clear_button(self, tr_elt, info):
        """ Write the configuration of the clear logs button of a process.
        This action must be sent to the relevant node. """
        namespec = info['namespec']
        self._write_process_button(tr_elt, 'clear_a_mid', info['identifier'], self.page_name,
                                   'clearlog', namespec, info['has_stdout'] or info['has_stderr'])

    def write_process_stdout_button(self, tr_elt, info):
        """ Write the configuration of the tail stdout button of a process.
        This action must be sent to the relevant node. """
        namespec = info['namespec']
        # no action requested. page name is enough
        self._write_process_button(tr_elt, 'tailout_a_mid', info['identifier'],
                                   STDOUT_PAGE % quote(namespec or ''), '', namespec, info['has_stdout'])

    def write_process_stderr_button(self, tr_elt, info):
        """ Write the configuration of the tail stderr button of a process.
        This action must be sent to the relevant node. """
        namespec = info['namespec']
        # no action requested. page name is enough
        self._write_process_button(tr_elt, 'tailerr_a_mid', info['identifier'],
                                   STDERR_PAGE % quote(namespec or ''), '', namespec, info['has_stderr'])

    def _write_process_button(self, tr_elt, elt_name: str, identifier: str, page: str, action: str, namespec: str,
                              active: bool = True):
        """ Write the configuration of a process button. """
        elt = tr_elt.findmeld(elt_name)
        if active:
            update_attrib(elt, 'class', 'button on')
            url = self.view_ctx.format_url(identifier, page, **{ACTION: action, NAMESPEC: namespec})
            elt.attributes(href=url)
        else:
            update_attrib(elt, 'class', 'button off')

    def write_common_process_table(self, table_elt):
        """ Hide MEM+CPU head+foot cells if statistics disabled"""
        if not self.has_process_statistics:
            for mid in ['mem_head_th_mid', 'cpu_head_th_mid',
                        'mem_foot_th_mid', 'cpu_foot_th_mid',
                        'mem_total_th_mid', 'cpu_total_th_mid']:
                elt = table_elt.findmeld(mid)
                if elt is not None:
                    elt.deparent()

    @staticmethod
    def write_common_state(tr_elt, info: Payload) -> None:
        """ Write the common part of a process state into a table. """
        # print state in cell
        td_elt = tr_elt.findmeld('state_td_mid')
        span_elt = td_elt.findmeld('state_span_mid')
        update_attrib(td_elt, 'class', info['gravity'])
        if info['has_crashed']:
            update_attrib(td_elt, 'class', 'crashed')
        if info['disabled']:
            update_attrib(td_elt, 'class', 'disabled')
        if len(info.get('running_identifiers', [])) > 1:
            update_attrib(span_elt, 'class', 'blink')
        span_elt.content(info['statename'])
        # print description
        tr_elt.findmeld('desc_td_mid').content(info['description'])

    def write_common_statistics(self, tr_elt, info: Payload) -> None:
        """ Write the common part of a process or application statistics into a table. """
        # print expected load
        tr_elt.findmeld('load_td_mid').content(f"{info['expected_load']}")
        # get data from statistics module iaw period selection
        self.write_common_process_cpu(tr_elt, info)
        self.write_common_process_mem(tr_elt, info)

    def write_common_process_status(self, tr_elt, info: Payload) -> None:
        """ Write the common part of a process status into a table. """
        # print common status
        self.write_common_state(tr_elt, info)
        self.write_common_statistics(tr_elt, info)
        # print process name
        # add break character only to processes belonging to an application
        process_name = info['process_name']
        if info.get('nb_items', 0) == 0:
            process_name = f'{SUB_SYMBOL} {process_name}'
        name_elt = tr_elt.findmeld('name_td_mid')
        # tail hyperlink depends on logfile availability
        namespec = info['namespec']
        # for application page, cannot state if logfile is there
        if info['has_stdout']:
            elt = name_elt.findmeld('name_a_mid')
            elt.content(process_name)
            url = self.view_ctx.format_url(info['identifier'], TAIL_PAGE,
                                           **{PROCESS: namespec, LIMIT: self.supvisors.options.tail_limit})
            elt.attributes(href=url, target="_blank")
        else:
            # overwrite a href entry
            name_elt.content(process_name)
        # manage actions iaw state
        self.write_process_start_button(tr_elt, info)
        self.write_process_stop_button(tr_elt, info)
        self.write_process_restart_button(tr_elt, info)
        # manage log actions
        self.write_process_clear_button(tr_elt, info)
        self.write_process_stdout_button(tr_elt, info)
        self.write_process_stderr_button(tr_elt, info)

    def write_detailed_process_cpu(self, stats_elt, proc_stats: Optional[ProcStatisticsInstance]) -> bool:
        """ Write the CPU part of the detailed process status.

        :param stats_elt: the element from which to search for.
        :param proc_stats: the process statistics.
        :return: True if process CPU statistics are valid.
        """
        if proc_stats:
            self._write_common_detailed_statistics(stats_elt, proc_stats.cpu, proc_stats.times,
                                                   'pcpuval_td_mid', 'pcpuavg_td_mid',
                                                   'pcpuslope_td_mid', 'pcpudev_td_mid')
            return True

    def write_detailed_process_mem(self, stats_elt, proc_stats: Optional[ProcStatisticsInstance]) -> bool:
        """ Write the MEM part of the detailed process status.

        :param stats_elt: the element from which to search for.
        :param proc_stats: the process statistics.
        :return: True if process Memory statistics are valid.
        """
        if proc_stats:
            self._write_common_detailed_statistics(stats_elt, proc_stats.mem, proc_stats.times,
                                                   'pmemval_td_mid', 'pmemavg_td_mid',
                                                   'pmemslope_td_mid', 'pmemdev_td_mid')
            return True

    def _write_common_detailed_statistics(self, ref_elt, stats, timeline, val_mid, avg_mid, slope_mid, dev_mid):
        """ Common rendering of statistics. """
        if len(stats) > 0:
            # get additional statistics
            avg, rate, (slope, _), dev = get_stats(timeline, stats)
            # set last value
            elt = ref_elt.findmeld(val_mid)
            if rate is not None and not math.isinf(rate):
                self.set_slope_class(elt, rate)
            elt.content(get_small_value(stats[-1]))
            # set mean value
            ref_elt.findmeld(avg_mid).content(get_small_value(avg))
            # adapt slope value iaw period selected
            if slope is not None:
                value = slope * self.view_ctx.period
                ref_elt.findmeld(slope_mid).content(get_small_value(value))
            # set standard deviation
            if dev is not None:
                ref_elt.findmeld(dev_mid).content(get_small_value(dev))

    def write_process_plots(self, proc_stats: ProcStatisticsInstance) -> bool:
        """ Write the CPU / Memory plots (only if matplotlib is installed) """
        try:
            from .plot import StatisticsPlot
            # build CPU image (if SOLARIS mode configured, CPU values have already been adjusted)
            cpu_img = StatisticsPlot(self.logger)
            cpu_img.add_timeline(proc_stats.times)
            cpu_img.add_plot('CPU', '%', proc_stats.cpu)
            cpu_img.export_image(process_cpu_img)
            # build Memory image
            mem_img = StatisticsPlot(self.logger)
            mem_img.add_timeline(proc_stats.times)
            mem_img.add_plot('MEM', '%', proc_stats.mem)
            mem_img.export_image(process_mem_img)
            return True
        except ImportError:
            # matplotlib not installed
            return False

    def write_process_statistics(self, root, info: Payload) -> None:
        """ Display detailed statistics about the selected process. """
        # update the statistics table
        stats_elt = root.findmeld('pstats_div_mid')
        # get data from statistics module iaw period selection
        namespec = info.get('namespec', None)
        if namespec:
            # set CPU/MEM statistics
            proc_stats: ProcStatisticsInstance = info['proc_stats']
            done_cpu = self.write_detailed_process_cpu(stats_elt, proc_stats)
            done_mem = self.write_detailed_process_mem(stats_elt, proc_stats)
            if done_cpu or done_mem:
                # set proces name
                stats_elt.findmeld('process_td_mid').content(namespec)
                # set node name and address
                supvisors_id: SupvisorsInstanceId = self.supvisors.mapper.instances[info['identifier']]
                stats_elt.findmeld('node_td_mid').content(supvisors_id.host_id)
                stats_elt.findmeld('ipaddress_td_mid').content(supvisors_id.ip_address)
                # write CPU / Memory plots
                if not self.write_process_plots(proc_stats):
                    # matplolib not installed: remove figure elements
                    for mid in ['cpuimage_fig_mid', 'memimage_fig_mid']:
                        stats_elt.findmeld(mid).replace('')
        else:
            # remove stats part if empty
            stats_elt.replace('')

    def handle_action(self) -> Optional[Type[NOT_DONE_YET]]:
        """ Handling of the actions requested from the Supvisors web pages.

        :return: NOT_DONE_YET if action is deferred, None otherwise
        """
        # check if any action is requested
        action = self.view_ctx.action
        if action:
            # trigger deferred action and wait
            namespec = self.view_ctx.namespec
            if not self.callback:
                self.callback = self.make_callback(namespec, action)
            # immediate check
            message = self.callback()
            if message is NOT_DONE_YET:
                return NOT_DONE_YET
            # post to write message
            if message is not None:
                self.view_ctx.store_message = format_gravity_message(message)

    def make_callback(self, namespec: str, action: str) -> Callable:
        """ Triggers processing iaw action requested.
        Subclasses will define what's to be done.

        :param namespec: the optional process namespec to which the action applies
        :param action: the action to perform
        :return: a callable for deferred result
        """
        raise NotImplementedError

    def multicall_rpc_action(self, args: PayloadList, success_msg: str) -> Callable:
        """ Generic wrapper for a System Supervisor RPC.

        :param args: the multiple RPC parameters
        :param success_msg: the message in case of success
        :return: a callable for deferred result
        """
        rpc_intf = self.supvisors.supervisor_data.system_rpc_interface
        return generic_rpc(rpc_intf, 'multicall', (args,), success_msg)

    def supervisor_rpc_action(self, rpc_name: str, args: tuple, success_msg: str) -> Callable:
        """ Generic wrapper for a Supervisor RPC.

        :param rpc_name: the RPC name in the Supervisor XML-RPC API
        :param args: the arguments of the RPC
        :param success_msg: the message in case of success
        :return: a callable for deferred result
        """
        rpc_intf = self.supvisors.supervisor_data.supervisor_rpc_interface
        return generic_rpc(rpc_intf, rpc_name, args, success_msg)

    def supvisors_rpc_action(self, rpc_name: str, args: tuple, success_msg: str) -> Callable:
        """ Generic wrapper for a Supvisors RPC.

        :param rpc_name: the RPC name in the Supvisors XML-RPC API
        :param args: the arguments of the RPC
        :param success_msg: the message in case of success
        :return: a callable for deferred result
        """
        rpc_intf = self.supvisors.supervisor_data.supvisors_rpc_interface
        return generic_rpc(rpc_intf, rpc_name, args, success_msg)

    @staticmethod
    def set_slope_class(elt, value: float) -> None:
        """ Set attribute class iaw positive or negative slope. """
        if abs(value) < .005:
            update_attrib(elt, 'class', 'stable')
        elif value > 0:
            update_attrib(elt, 'class', 'increase')
        else:
            update_attrib(elt, 'class', 'decrease')



# ======================================================================
# Copyright 2023 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import time
from socket import gethostname, gethostbyaddr

import pytest

from supvisors.internal_com.multicast import *


@pytest.fixture
def discovery(supvisors):
    """ Create the MulticastSender instance. """
    # set the options
    supvisors.options.multicast_group = '239.0.0.1', 7777
    supvisors.options.multicast_ttl = 1
    # create the instance
    disco = SupvisorsDiscovery(supvisors)
    # wait for thread to start
    time.sleep(1)
    assert disco.mc_receiver.is_alive()
    yield disco
    disco.stop()


def test_multicast(supvisors, discovery):
    """ Test the Supvisors Multicast in one single test. """
    hostname, aliases, ip_addresses = gethostbyaddr(gethostname())
    # send event
    payload = {'when': 1234, 'identifier': '10.0.0.1:10000', 'nick_identifier': 'rocky51',
               'http_port': 10000, 'ip_addresses': ip_addresses}
    discovery.mc_sender.send_discovery_event(payload)
    # check output
    time.sleep(1.0)
    call_args_list = supvisors.rpc_handler.push_notification.call_args_list
    assert len(call_args_list) == 1
    message = call_args_list[0][0][0]
    assert type(message) is tuple
    assert len(message) == 2
    source, event_body = message
    assert source[0] == '10.0.0.1:10000'
    assert source[1] == 'rocky51'
    assert type(source[2]) is tuple
    assert len(source[2]) == 2
    assert source[2][1] == 10000
    assert event_body == (NotificationHeaders.DISCOVERY.value, ())
    supvisors.rpc_handler.push_notification.reset_mock()
    # send event that does not fit
    payload = {'when': 1234, 'identifier': '10.0.0.1:10000', 'nick_identifier': 'rocky51',
               'http_port': 10000, 'ip_addresses': []}
    discovery.mc_sender.send_discovery_event(payload)
    assert len(call_args_list) == 1
    message = call_args_list[0][0][0]
    assert type(message) is tuple
    assert len(message) == 2
    source, event_body = message
    assert source[0] == '10.0.0.1:10000'
    assert source[1] == 'rocky51'
    assert type(source[2]) is tuple
    assert len(source[2]) == 2
    assert source[2][1] == 10000
    assert event_body == (NotificationHeaders.DISCOVERY.value, ())


# testing exception cases (by line number)
def test_emitter_send_exception(discovery):
    """ Test the sendto exception of the MulticastSender.
    The aim is to hit the lines 64-66 in MulticastSender.emit_message.
    Checked ok with debugger.
    """
    discovery.stop()
    discovery.mc_sender.send_discovery_event({})


def test_receiver_bind_exception(supvisors):
    """ Test the bind exception of the MulticastReceiver (use wrong IP).
    The aim is to hit the lines 131-134 in MulticastReceiver.open_multicast.
    Checked ok with debugger.
    """
    mc_receiver = MulticastReceiver(('localhost', -1), None, None, supvisors.logger)
    assert mc_receiver.socket is None
    mc_receiver.open_multicast()
    assert mc_receiver.socket is None


def test_receiver_membership_exception(supvisors):
    """ Test the add membership exception of the MulticastReceiver (use wrong IP).
    The aim is to hit the lines 145-148 in MulticastReceiver.open_multicast.
    Checked ok with debugger.
    """
    # test illegal address
    mc_receiver = MulticastReceiver(('localhost', 7777), 'localhost', None, supvisors.logger)
    assert mc_receiver.socket is None
    mc_receiver.open_multicast()
    assert mc_receiver.socket is None
    # test wrong interface
    mc_receiver = MulticastReceiver(('239.0.0.1', 7777), '10.0.0', None, supvisors.logger)
    assert mc_receiver.socket is None
    mc_receiver.open_multicast()
    assert mc_receiver.socket is None



# ======================================================================
# Copyright 2016 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

from supvisors.internal_com.mapper import SupvisorsInstanceId
from supvisors.statscompiler import HostStatisticsInstance, InterfaceHistoryStats
from .viewcontext import *
from .viewimage import host_cpu_img, host_mem_img, host_net_io_img, host_disk_io_img, host_disk_usage_img
from .viewinstance import SupvisorsInstanceView
from .webutils import *


class HostInstanceView(SupvisorsInstanceView):
    """ View renderer of the Host section of the Supvisors Instance page. """

    def __init__(self, context):
        """ Call of the superclass constructors. """
        SupvisorsInstanceView.__init__(self, context, HOST_INSTANCE_PAGE)

    def write_options(self, header_elt):
        """ Write configured periods for host statistics. """
        # in the current design, this page should not be accessed if the host statistics are not enabled
        # so no need to hide the period buttons in this case
        self.write_periods(header_elt)
        # always allow to go back to process view
        self.write_view_switch(header_elt)

    def write_view_switch(self, header_elt):
        """ Configure the statistics view buttons. """
        # update process button
        elt = header_elt.findmeld('process_view_a_mid')
        url = self.view_ctx.format_url('', PROC_INSTANCE_PAGE)
        elt.attributes(href=url)
        # update host button
        elt = header_elt.findmeld('host_view_a_mid')
        elt.content(f'{self.sup_ctx.local_status.supvisors_id.host_id}')

    def write_contents(self, contents_elt):
        """ Rendering of tables and figures for address statistics. """
        # get node characteristics
        info = self.view_ctx.get_node_characteristics()
        if info:
            self.write_node_characteristics(contents_elt, info)
        # get data from statistics module iaw period selection
        stats_instance: HostStatisticsInstance = self.view_ctx.get_instance_stats()
        if stats_instance and stats_instance.times:
            self.write_processor_statistics(contents_elt, stats_instance.cpu, stats_instance.times)
            self.write_memory_statistics(contents_elt, stats_instance.mem, stats_instance.times)
            self.write_network_io_statistics(contents_elt, stats_instance.net_io)
            self.write_disk_statistics(contents_elt, stats_instance)
            # write CPU / Memory / Network plots
            try:
                self._write_cpu_image(stats_instance.cpu, stats_instance.times)
                self._write_mem_image(stats_instance.mem, stats_instance.times)
                # other statistics include their own timeline
                self._write_net_io_image(stats_instance.net_io)
                self._write_disk_image(stats_instance)
            except ImportError:
                # matplotlib not installed: remove figure elements
                for mid in ['cpu_image_fig_mid', 'mem_image_fig_mid', 'net_io_image_fig_mid',
                            'disk_io_image_fig_mid', 'disk_usage_image_fig_mid']:
                    contents_elt.findmeld(mid).replace('')

    def write_node_characteristics(self, contents_elt, info):
        """ Rendering of the node characteristics. """
        # write Node section
        supvisors_id: SupvisorsInstanceId = self.sup_ctx.local_status.supvisors_id
        contents_elt.findmeld('node_td_mid').content(supvisors_id.host_id)
        contents_elt.findmeld('ipaddress_td_mid').content(supvisors_id.ip_address)
        # write Processor section
        contents_elt.findmeld('cpu_count_td_mid').content(f'{info.nb_core_physical} / {info.nb_core_logical}')
        contents_elt.findmeld('cpu_freq_td_mid').content(info.frequency)
        # write Memory section
        contents_elt.findmeld('physical_mem_td_mid').content(info.physical_memory)

    def _write_processor_single_title(self, tr_elt, selected_cpu_id, cpu_id):
        """ Rendering of the title of a single core. """
        elt = tr_elt.findmeld('cpu_num_a_mid')
        if selected_cpu_id == cpu_id:
            elt.attrib['class'] = 'button off active'
        else:
            url = self.view_ctx.format_url('', self.page_name, **{CPU: cpu_id})
            elt.attributes(href=url)
        cpu_id_string = self.view_ctx.cpu_id_to_string(cpu_id)
        elt.content('cpu#%s' % cpu_id_string)

    def _write_processor_single_statistics(self, tr_elt, single_cpu_stats, timeline):
        """ Rendering of the processor statistics for a single core. """
        self._write_common_detailed_statistics(tr_elt, single_cpu_stats, timeline,
                                               'cpu_val_td_mid', 'cpu_avg_td_mid',
                                               'cpu_slope_td_mid', 'cpu_dev_td_mid')

    def write_processor_statistics(self, root, cpu_stats, timeline):
        """ Rendering of the processor statistics. """
        selected_cpu_id = self.view_ctx.cpu_id
        iterator = root.findmeld('cpu_tr_mid').repeat(cpu_stats)
        shaded_tr = False
        for cpu_id, (tr_elt, single_cpu_stats) in enumerate(iterator):
            self._write_processor_single_title(tr_elt, selected_cpu_id, cpu_id)
            self._write_processor_single_statistics(tr_elt, single_cpu_stats, timeline)
            # set row background and invert
            apply_shade(tr_elt, shaded_tr)
            shaded_tr = not shaded_tr

    def write_memory_statistics(self, root, mem_stats, timeline):
        """ Rendering of the memory statistics. """
        self._write_common_detailed_statistics(root, mem_stats, timeline,
                                               'mem_val_td_mid', 'mem_avg_td_mid',
                                               'mem_slope_td_mid', 'mem_dev_td_mid')

    def _write_network_io_single_title(self, tr_elt, selected_nic, nic, rowspan, shaded_tr):
        """ Rendering of the title column of the network statistics. """
        # set interface cell rowspan
        elt = tr_elt.findmeld('net_io_nic_td_mid')
        if rowspan:
            elt.attrib['rowspan'] = '2'
            # apply shade logic to td element too for background-image to work
            apply_shade(elt, shaded_tr)
            # set interface name on a/href elt
            elt = elt.findmeld('net_io_nic_a_mid')
            elt.content(nic)
            if selected_nic == nic:
                elt.attrib['class'] = 'button off active'
            else:
                url = self.view_ctx.format_url('', self.page_name, **{NIC: nic})
                elt.attributes(href=url)
        else:
            elt.replace('')

    def _write_network_io_single_statistics(self, tr_elt, single_io_stats, timeline, rowspan):
        """ Rendering of the statistics columns of the network statistics. """
        elt = tr_elt.findmeld('net_io_rxtx_td_mid')
        elt.content('Rx' if rowspan else 'Tx')
        # calculate and write statistics
        self._write_common_detailed_statistics(tr_elt, single_io_stats, timeline,
                                               'net_io_val_td_mid', 'net_io_avg_td_mid',
                                               'net_io_slope_td_mid', 'net_io_dev_td_mid')

    def write_network_io_statistics(self, root, io_stats):
        """ Rendering of the network IO statistics. """
        selected_nic = self.view_ctx.nic_name
        # display io statistics
        flatten_io_stats = []
        for nic, (uptimes, [recv_stats, sent_stats]) in io_stats.items():
            flatten_io_stats.append((nic, uptimes, recv_stats))
            flatten_io_stats.append((nic, uptimes, sent_stats))
        iterator = root.findmeld('net_io_tr_mid').repeat(flatten_io_stats)
        rowspan, shaded_tr = True, False
        for tr_elt, (nic, uptimes, single_io_stats) in iterator:
            # set row background
            apply_shade(tr_elt, shaded_tr)
            # set interface cell rowspan
            self._write_network_io_single_title(tr_elt, selected_nic, nic, rowspan, shaded_tr)
            # set interface direction
            self._write_network_io_single_statistics(tr_elt, single_io_stats, uptimes, rowspan)
            if not rowspan:
                shaded_tr = not shaded_tr
            rowspan = not rowspan

    def _write_disk_io_single_title(self, tr_elt, selected_device, device, rowspan, shaded_tr):
        """ Rendering of the title column of the disk IO statistics. """
        # set device cell rowspan
        elt = tr_elt.findmeld('disk_io_device_td_mid')
        if rowspan:
            elt.attrib['rowspan'] = '2'
            # apply shade logic to td element too for background-image to work
            apply_shade(elt, shaded_tr)
            # set device name on a/href elt
            elt = elt.findmeld('disk_io_device_a_mid')
            elt.content(device)
            if selected_device == device:
                elt.attrib['class'] = 'button off active'
            else:
                url = self.view_ctx.format_url('', self.page_name, **{DEVICE: device})
                elt.attributes(href=url)
        else:
            elt.replace('')

    def _write_disk_io_single_statistics(self, tr_elt, single_io_stats, timeline, rowspan):
        """ Rendering of the statistics columns of the disk IO statistics. """
        elt = tr_elt.findmeld('disk_io_rw_td_mid')
        elt.content('R' if rowspan else 'W')
        # calculate and write statistics
        self._write_common_detailed_statistics(tr_elt, single_io_stats, timeline,
                                               'disk_io_val_td_mid', 'disk_io_avg_td_mid',
                                               'disk_io_slope_td_mid', 'disk_io_dev_td_mid')

    def write_disk_io_statistics(self, contents_elt, io_stats):
        """ Rendering of the disk IO statistics. """
        selected_device = self.view_ctx.device
        # display io device statistics
        flatten_io_stats = []
        for device, (uptimes, [read_stats, write_stats]) in io_stats.items():
            flatten_io_stats.append((device, uptimes, read_stats))
            flatten_io_stats.append((device, uptimes, write_stats))
        iterator = contents_elt.findmeld('disk_io_tr_mid').repeat(flatten_io_stats)
        rowspan, shaded_tr = True, False
        for tr_elt, (device, uptimes, single_io_stats) in iterator:
            # set row background
            apply_shade(tr_elt, shaded_tr)
            # set device cell rowspan
            self._write_disk_io_single_title(tr_elt, selected_device, device, rowspan, shaded_tr)
            # set device direction
            self._write_disk_io_single_statistics(tr_elt, single_io_stats, uptimes, rowspan)
            if not rowspan:
                shaded_tr = not shaded_tr
            rowspan = not rowspan

    def _write_disk_usage_single_title(self, tr_elt, selected_partition, partition):
        """ Rendering of the title column of the network statistics. """
        # set partition name on a/href elt
        elt = tr_elt.findmeld('disk_usage_partition_a_mid')
        elt.content(partition)
        if selected_partition == partition:
            elt.attrib['class'] = 'button off active'
        else:
            url = self.view_ctx.format_url('', self.page_name, **{PARTITION: partition})
            elt.attributes(href=url)

    def _write_disk_usage_single_statistics(self, tr_elt, usage_stats, timeline):
        """ Rendering of the statistics columns of the disk usage statistics. """
        self._write_common_detailed_statistics(tr_elt, usage_stats, timeline,
                                               'disk_usage_val_td_mid', 'disk_usage_avg_td_mid',
                                               'disk_usage_slope_td_mid', 'disk_usage_dev_td_mid')

    def write_disk_usage_statistics(self, contents_elt, usage_stats: InterfaceHistoryStats):
        """ Rendering of the network statistics. """
        selected_partition = self.view_ctx.partition
        iterator = contents_elt.findmeld('disk_usage_tr_mid').repeat(usage_stats.items())
        shaded_tr = False
        for tr_elt, (partition, (uptimes, [single_usage_stats])) in iterator:
            self._write_disk_usage_single_title(tr_elt, selected_partition, partition)
            self._write_disk_usage_single_statistics(tr_elt, single_usage_stats, uptimes)
            # set row background and invert
            apply_shade(tr_elt, shaded_tr)
            shaded_tr = not shaded_tr

    def write_disk_statistics(self, contents_elt, stats_instance: HostStatisticsInstance):
        """ Rendering the relevant Disk statistics depending on the display choice. """
        if self.view_ctx.disk_stats == 'usage':
            # update Disk usage stats and hide Disk IO stats
            self.write_disk_usage_statistics(contents_elt, stats_instance.disk_usage)
            contents_elt.findmeld('disk_io_div_mid').replace('')
            # update the disk stats selector
            url = self.view_ctx.format_url('', self.page_name, **{DISK_STATS: 'io'})
            contents_elt.findmeld('disk_usage_view_mid').attributes(href=url)
        else:
            # update Disk IO stats and hide Disk usage stats
            self.write_disk_io_statistics(contents_elt, stats_instance.disk_io)
            contents_elt.findmeld('disk_usage_div_mid').replace('')
            # update the disk stats selector
            url = self.view_ctx.format_url('', self.page_name, **{DISK_STATS: 'usage'})
            contents_elt.findmeld('disk_io_view_mid').attributes(href=url)

    def _write_cpu_image(self, cpu_stats, timeline):
        """ Write CPU data into the dedicated buffer. """
        from supvisors.web.plot import StatisticsPlot
        # get CPU data
        cpu_id = self.view_ctx.cpu_id
        cpu_id_string = self.view_ctx.cpu_id_to_string(cpu_id)
        cpu_data = cpu_stats[cpu_id]
        # build image from data
        plt = StatisticsPlot(self.logger)
        plt.add_timeline(timeline)
        plt.add_plot(f'CPU #{cpu_id_string}', '%', cpu_data)
        plt.export_image(host_cpu_img)

    def _write_mem_image(self, mem_stats, timeline):
        """ Write MEM data into the dedicated buffer. """
        # build image from data
        from supvisors.web.plot import StatisticsPlot
        plt = StatisticsPlot(self.logger)
        plt.add_timeline(timeline)
        plt.add_plot('MEM', '%', mem_stats)
        plt.export_image(host_mem_img)

    def _write_net_io_image(self, net_io_stats):
        """ Write Network IO data into the dedicated buffer. """
        from supvisors.web.plot import StatisticsPlot
        # get IO data
        nic_name = self.view_ctx.nic_name
        if nic_name:
            timeline, [recv_data, sent_data] = net_io_stats[nic_name]
            # build image from data
            plt = StatisticsPlot(self.logger)
            plt.add_timeline(timeline)
            plt.add_plot(f'{nic_name} received', 'kbits/s', recv_data)
            plt.add_plot(f'{nic_name} sent', 'kbits/s', sent_data)
            plt.export_image(host_net_io_img)

    def _write_disk_image(self, stats_instance: HostStatisticsInstance):
        """ Update the relevant Disk image depending on the display choice. """
        if self.view_ctx.disk_stats == 'usage':
            # update Disk usage image
            self._write_disk_usage_image(stats_instance.disk_usage)
        else:
            # update Disk IO image
            self._write_disk_io_image(stats_instance.disk_io)

    def _write_disk_io_image(self, disk_io_stats):
        """ Write Disk IO data into the dedicated buffer. """
        from supvisors.web.plot import StatisticsPlot
        # get IO data
        device_name = self.view_ctx.device
        if device_name:
            timeline, [read_data, write_data] = disk_io_stats[device_name]
            # build image from data
            plt = StatisticsPlot(self.logger)
            plt.add_timeline(timeline)
            plt.add_plot(f'{device_name} read', 'kbits/s', read_data)
            plt.add_plot(f'{device_name} written', 'kbits/s', write_data)
            plt.export_image(host_disk_io_img)

    def _write_disk_usage_image(self, usage_stats):
        """ Write Disk usage data into the dedicated buffer. """
        from supvisors.web.plot import StatisticsPlot
        # get usage data
        partition = self.view_ctx.partition
        if partition:
            timeline, [usage_stats] = usage_stats[partition]
            # build image from data
            plt = StatisticsPlot(self.logger)
            plt.add_timeline(timeline)
            plt.add_plot(f'{partition}', '%', usage_stats)
            plt.export_image(host_disk_usage_img)



# ======================================================================
# Copyright 2016 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import errno
import socket
import sys
from types import MethodType

from supervisor import supervisorctl
from supervisor import xmlrpc
from supervisor.compat import as_string, xmlrpclib
from supervisor.loggers import getLevelNumByDescription
from supervisor.options import ClientOptions, make_namespec, split_namespec
from supervisor.states import ProcessStates, getProcessStateDescription
from supervisor.supervisorctl import Controller, ControllerPluginBase, LSBInitExitStatuses

from . import __version__
from .plugin import expand_faults
from .rpcinterface import RPCInterface
from .ttypes import ConciliationStrategies, StartingStrategies, SupvisorsInstanceStates, PayloadList
from .utils import simple_localtime


def _startresult(self, result):
    """ This method replaces the Supervisor method in DefaultControllerPlugin.

    :param result: the command result
    :return: the result to output
    """
    if result['status'] == xmlrpc.Faults.DISABLED:
        name = make_namespec(result['group'], result['name'])
        return f'{name}: ERROR disabled'
    return self._startresult_ref(result)


class ControllerPlugin(ControllerPluginBase):
    """ The ControllerPlugin is the implementation of the Supvisors plugin that is embodied in the supervisorctl
    command. """

    def __init__(self, controller):
        super().__init__(controller)
        # update Supervisor Fault definition
        expand_faults()
        # patch supervisor plugin to manage new Faults.DISABLED code in RPCError
        supervisor_plugin = controller.options.plugins[0]
        if not hasattr(supervisor_plugin, '_startresult_ref'):
            supervisor_plugin._startresult_ref = supervisor_plugin._startresult
            supervisor_plugin._startresult = MethodType(_startresult, supervisor_plugin)

    def get_server_proxy(self, server_url: str, namespace: str):
        """ Return a proxy on a remote supervisor instance. """
        proxy = xmlrpclib.ServerProxy('http://127.0.0.1',
                                      xmlrpc.SupervisorTransport(self.ctl.options.username,
                                                                 self.ctl.options.password,
                                                                 server_url))
        return getattr(proxy, namespace)

    def supvisors(self) -> RPCInterface:
        """ Get a proxy to the Supvisors RPC interface. """
        return self.ctl.get_server_proxy('supvisors')

    def get_running_instances(self):
        """ Return the URL of the Supvisors running instances. """
        try:
            # get everything at once instead of doing multiple requests
            info_list = self.supvisors().get_all_instances_info()
        except xmlrpclib.Fault as e:
            self.ctl.output(f'ERROR ({e.faultString})')
            self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
            info_list = []
        return {info['identifier']: 'http://%s:%d' % (info['node_name'], info['port'])
                for info in info_list
                if info['statecode'] in [SupvisorsInstanceStates.CHECKED.value, SupvisorsInstanceStates.RUNNING.value]}

    def do_sversion(self, _):
        """ Command to get the Supvisors API version. """
        if self._upcheck():
            try:
                version = self.supvisors().get_api_version()
            except xmlrpclib.Fault as e:
                self.ctl.output(f'ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
            else:
                self.ctl.output(version)

    def help_sversion(self):
        """ Print the help of the sversion command. """
        self.ctl.output('sversion\t\t\tGet the API version of Supvisors.')

    def do_sstate(self, _):
        """ Command to get the Supvisors state. """
        if self._upcheck():
            try:
                state_modes = self.supvisors().get_supvisors_state()
            except xmlrpclib.Fault as e:
                self.ctl.output(f'ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
            else:
                max_master = ControllerPlugin.max_template([state_modes], 'master_identifier', 'Master')
                max_starting = ControllerPlugin.max_template([state_modes], 'starting_jobs', 'Starting')
                max_stopping = ControllerPlugin.max_template([state_modes], 'stopping_jobs', 'Stopping')
                template = (f'%(state)-16s%(discovery)-11s%(master)-{max_master}s%(starting)-{max_starting}s'
                            f'%(stopping)-{max_stopping}s')
                # print title
                payload = {'state': 'State', 'discovery': 'Discovery', 'master': 'Master',
                           'starting': 'Starting', 'stopping': 'Stopping'}
                self.ctl.output(template % payload)
                # print data
                line = template % {'state': state_modes['fsm_statename'],
                                   'discovery': state_modes['discovery_mode'],
                                   'master': state_modes['master_identifier'],
                                   'starting': state_modes['starting_jobs'],
                                   'stopping': state_modes['stopping_jobs']}
                self.ctl.output(line)

    def help_sstate(self):
        """ Print the help of the sstate command. """
        self.ctl.output('sstate\t\t\t\tGet the Supvisors state.')

    def do_master(self, _):
        """ Command to get the Master Supvisors instance. """
        if self._upcheck():
            try:
                identifier = self.supvisors().get_master_identifier()
            except xmlrpclib.Fault as e:
                self.ctl.output(f'ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
            else:
                self.ctl.output(identifier)

    def help_master(self):
        """ Print the help of the master command. """
        self.ctl.output('master\t\t\t\tGet the Master Supvisors instance.')

    def do_strategies(self, _):
        """ Command to get the Supvisors strategies. """
        if self._upcheck():
            try:
                strategies = self.supvisors().get_strategies()
            except xmlrpclib.Fault as e:
                self.ctl.output(f'ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
            else:
                self.ctl.output(f"Auto-fencing: {strategies['auto-fencing']}")
                self.ctl.output(f"Conciliation: {strategies['conciliation']}")
                self.ctl.output(f"Starting:     {strategies['starting']}")

    def help_strategies(self):
        """ Print the help of the strategies command."""
        self.ctl.output('strategies\t\t\tGet the Supvisors strategies.')

    def do_stats_status(self, _):
        """ Command to get the Supvisors statistics collection status. """
        if self._upcheck():
            try:
                status = self.supvisors().get_statistics_status()
            except xmlrpclib.Fault as e:
                self.ctl.output(f'ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
            else:
                self.ctl.output(f"Collecting Host statistics:    {status['host_stats']}")
                self.ctl.output(f"Collecting Process statistics: {status['process_stats']}")
                self.ctl.output(f"Collecting statistics period:  {status['collecting_period']}")

    def help_stats_status(self):
        """ Print the help of the Supvisors statistics collection status command."""
        self.ctl.output('stats_status\t\t\tGet the Supvisors statistics collection status.')

    def do_instance_status(self, arg):
        """ Command to get the status of instances known to Supvisors. """
        if self._upcheck():
            try:
                # get everything at once instead of doing multiple requests
                info_list = self.supvisors().get_all_instances_info()
            except xmlrpclib.Fault as e:
                self.ctl.output(f'ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
            else:
                # create template. identifier has variable length
                max_nick_identifiers = ControllerPlugin.max_template(info_list, 'nick_identifier', 'Nickname')
                max_identifiers = ControllerPlugin.max_template(info_list, 'identifier', 'Supvisors identifier')
                max_hosts = ControllerPlugin.max_template(info_list, 'node_name', 'Host')
                max_masters = ControllerPlugin.max_template(info_list, 'master_identifier', 'Master')
                template = (f'%(nick_identifier)-{max_nick_identifiers}s%(identifier)-{max_identifiers}s'
                            f'%(node_name)-{max_hosts}s%(port)-7s'
                            '%(state)-11s%(discovery)-11s%(load)-6s%(ltime)-10s%(counter)-9s'
                            '%(proc_failure)-11s'
                            f'%(fsm_state)-16s%(discovery)-11s%(master)-{max_masters}s'
                            '%(starting)-10s%(stopping)-10s')
                # print title
                payload = {'nick_identifier': 'Nickname', 'identifier': 'Supvisors identifier',
                           'node_name': 'Host', 'port': 'Port',
                           'state': 'State', 'discovery': 'Discovery',
                           'load': 'Load', 'ltime': 'Time', 'counter': 'Counter',
                           'proc_failure': 'Processes',
                           'fsm_state': 'FSM', 'master': 'Master',
                           'starting': 'Starting', 'stopping': 'Stopping'}
                self.ctl.output(template % payload)
                # check request args
                identifiers = arg.split()
                output_all = not identifiers or 'all' in identifiers
                # print filtered payloads
                for info in info_list:
                    if output_all or info['identifier'] in identifiers:
                        payload = {'nick_identifier': info['nick_identifier'],
                                   'identifier': info['identifier'],
                                   'node_name': info['node_name'], 'port': info['port'],
                                   'state': info['statename'],
                                   'discovery': info['discovery_mode'],
                                   'load': f"{info['loading']}%",
                                   'ltime': simple_localtime(info['local_time']),
                                   'counter': info['remote_sequence_counter'],
                                   'proc_failure': 'error' if info['process_failure'] else 'ok',
                                   'fsm_state': info['fsm_statename'],
                                   'master': info['master_identifier'],
                                   'starting': info['starting_jobs'],
                                   'stopping': info['stopping_jobs']}
                        self.ctl.output(template % payload)

    def help_instance_status(self):
        """ Print the help of the instance_status command."""
        self.ctl.output('instance_status <identifier>\t\t\tGet the status of the Supvisors instance.')
        self.ctl.output('instance_status <identifier> <identifier>\tGet the status for multiple Supvisors instances')
        self.ctl.output('instance_status\t\t\t\t\tGet the status of all Supvisors instances.')

    @staticmethod
    def max_template(payloads: PayloadList, item: str, title: str):
        """ Return the length of the item column taking into account the title and all its elements.

        :param payloads: the list of elements
        :param item: the item to display
        :param title: the title of the column
        :return: the length of the column
        """
        max_item = len(title)
        # WARN: item value is not always a string
        values = [len(str(info[item])) for info in payloads if item in info]
        return max(max_item, max(values) if values else 0) + 2

    def do_application_info(self, arg):
        """ Command to get information about applications known to Supvisors. """
        if self._upcheck():
            try:
                # get everything at once instead of doing multiple requests
                info_list = self.supvisors().get_all_applications_info()
            except xmlrpclib.Fault as e:
                self.ctl.output(f'ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
            else:
                # create template. identifier has variable length
                max_appli = ControllerPlugin.max_template(info_list, 'application_name', 'Application')
                template = '%(name)-{}s%(state)-10s%(major_failure)-7s%(minor_failure)-7s'.format(max_appli)
                # print title
                payload = {'name': 'Application', 'state': 'State', 'major_failure': 'Major', 'minor_failure': 'Minor'}
                self.ctl.output(template % payload)
                # check request args
                applications = arg.split()
                output_all = not applications or "all" in applications
                # print filtered payloads
                for info in info_list:
                    if output_all or info['application_name'] in applications:
                        payload = {'name': info['application_name'], 'state': info['statename'],
                                   'major_failure': info['major_failure'], 'minor_failure': info['minor_failure']}
                        self.ctl.output(template % payload)

    def help_application_info(self):
        """ Print the help of the application_info command."""
        self.ctl.output("application_info <appli>\t\t"
                        "Get the status of application named appli.")
        self.ctl.output("application_info <appli> <appli>\t"
                        "Get the status for multiple named application")
        self.ctl.output("application_info\t\t\t"
                        "Get the status of all applications.")

    def do_application_rules(self, arg):
        """ Command to get the application rules handled by Supvisors. """
        if self._upcheck():
            applications = arg.split()
            if not applications or 'all' in applications:
                try:
                    applications = [application_info['application_name']
                                    for application_info in self.supvisors().get_all_applications_info()]
                except xmlrpclib.Fault as e:
                    self.ctl.output(f'ERROR ({e.faultString})')
                    self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                    applications = []
            rules_list = []
            for application in applications:
                try:
                    rules = self.supvisors().get_application_rules(application)
                except xmlrpclib.Fault as e:
                    self.ctl.output(f'{application}: ERROR ({e.faultString})')
                    self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                else:
                    rules_list.append(rules)
            # print results
            if rules_list:
                max_appli = ControllerPlugin.max_template(rules_list, 'application_name', 'Application')
                max_identifiers = ControllerPlugin.max_template(rules_list, 'identifiers', 'Supervisor')
                # print title
                template = (f'%(appli)-{max_appli}s%(managed)-9s%(distribution)-17s%(identifiers)-{max_identifiers}s'
                            '%(start_seq)-7s%(stop_seq)-7s%(starting_strategy)-18s%(starting_failure_strategy)-18s'
                            '%(running_failure_strategy)s')
                title = {'appli': 'Application', 'managed': 'Managed', 'distribution': 'Distribution',
                         'identifiers': 'Supervisor', 'start_seq': 'Start', 'stop_seq': 'Stop',
                         'starting_strategy': 'Starting', 'starting_failure_strategy': 'Starting_Failure',
                         'running_failure_strategy': 'Running_Failure'}
                self.ctl.output(template % title)
                # print rules
                for rules in rules_list:
                    if rules['managed']:
                        payload = {'appli': rules['application_name'], 'managed': True,
                                   'distribution': rules['distribution'],
                                   'identifiers': rules.get('identifiers', 'n/a'),
                                   'start_seq': rules['start_sequence'], 'stop_seq': rules['stop_sequence'],
                                   'starting_strategy': rules['starting_strategy'],
                                   'starting_failure_strategy': rules['starting_failure_strategy'],
                                   'running_failure_strategy': rules['running_failure_strategy']}
                    else:
                        payload = {'appli': rules['application_name'], 'managed': False,
                                   'distribution': 'n/a', 'identifiers': 'n/a',
                                   'start_seq': 'n/a', 'stop_seq': 'n/a',
                                   'starting_strategy': 'n/a',
                                   'starting_failure_strategy': 'n/a',
                                   'running_failure_strategy': 'n/a'}
                    self.ctl.output(template % payload)

    def help_application_rules(self):
        """ Print the help of the application_rules command."""
        self.ctl.output('application_rules <appli>\t\t\tGet the rules of the application named appli.')
        self.ctl.output('application_rules <appli> <appli>\t\tGet the rules for multiple named applications')
        self.ctl.output('application_rules\t\t\t\tGet the rules of all applications.')

    def do_sstatus(self, arg):
        """ Command to get information about processes known to Supvisors. """
        if self._upcheck():
            processes = arg.split()
            if not processes or 'all' in processes:
                try:
                    info_list = self.supvisors().get_all_process_info()
                except xmlrpclib.Fault as e:
                    self.ctl.output(f'ERROR ({e.faultString})')
                    self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                    info_list = []
            else:
                info_list = []
                for process in processes:
                    try:
                        info = self.supvisors().get_process_info(process)
                    except xmlrpclib.Fault as e:
                        self.ctl.output(f'{process}: ERROR ({e.faultString})')
                        self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                    else:
                        info_list.extend(info)
            # print results
            if info_list:
                max_appli = ControllerPlugin.max_template(info_list, 'application_name', 'Application')
                max_proc = ControllerPlugin.max_template(info_list, 'process_name', 'Process')
                # print title
                template = f'%(appli)-{max_appli}s%(proc)-{max_proc}s%(state)-12s%(expected)-10s%(identifiers)s'
                title = {'appli': 'Application', 'proc': 'Process', 'state': 'State',
                         'expected': 'Expected', 'identifiers': 'Supervisor'}
                self.ctl.output(template % title)
                # print process status
                for info in info_list:
                    expected = info['expected_exit'] if info['statecode'] == ProcessStates.EXITED else 'n/a'
                    payload = {'appli': info['application_name'],
                               'proc': info['process_name'],
                               'state': info['statename'],
                               'expected': expected,
                               'identifiers': info['identifiers']}
                    self.ctl.output(template % payload)

    def help_sstatus(self):
        """ Print the help of the sstatus command."""
        self.ctl.output('sstatus <proc>\t\t\t\tGet the status of the process named proc.')
        self.ctl.output('sstatus <appli>:*\t\t\tGet the process status of application named appli.')
        self.ctl.output('sstatus <proc> <proc>\t\t\tGet the status for multiple named processes')
        self.ctl.output('sstatus\t\t\t\t\tGet the status of all processes.')

    def do_local_status(self, arg):
        """ Command to get a subset of information about processes, using Supervisor's getProcessInfo,
        complemented by extra arguments.
        This is the minimal information required by Supvisors. """
        if self._upcheck():
            # get all information
            try:
                info_list = self.supvisors().get_all_local_process_info()
            except xmlrpclib.Fault as e:
                self.ctl.output(f'ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                info_list = []
            # filter information iaw arguments
            processes = arg.split()
            if not processes or 'all' in processes:
                match_list = info_list
            else:
                match_list = []
                for process in processes:
                    group_name, process_name = split_namespec(process)
                    for info in info_list:
                        matches = info['group'] == group_name
                        if process_name:
                            matches &= info['name'] == process_name
                        if matches:
                            match_list.append(info)
            # print results
            if match_list:
                max_appli = ControllerPlugin.max_template(match_list, 'group', 'Application')
                max_proc = ControllerPlugin.max_template(match_list, 'name', 'Process')
                template = (f'%(appli)-{max_appli}s%(proc)-{max_proc}s%(disabled)-10s%(state)-12s%(start)-12s'
                            '%(now)-12s%(pid)-8s%(args)s')
                # print title
                payload = {'appli': 'Application', 'proc': 'Process', 'disabled': 'Disabled', 'state': 'State',
                           'start': 'Start', 'now': 'Now', 'pid': 'PID', 'args': 'Extra args'}
                self.ctl.output(template % payload)
                # print filtered payloads (some attributes serve only Supvisors internal purpose)
                for info in match_list:
                    start_time = simple_localtime(info['start']) if info['start'] else 0
                    now_time = simple_localtime(info['now']) if info['now'] else 0
                    payload = {'appli': info['group'], 'proc': info['name'], 'disabled': info['disabled'],
                               'state': getProcessStateDescription(info['state']),
                               'start': start_time, 'now': now_time, 'pid': info['pid'],
                               'args': info['extra_args']}
                    self.ctl.output(template % payload)

    def help_local_status(self):
        """ Print the help of the local_status command."""
        self.ctl.output('local_status <proc>\t\t\tGet the local status of the process named proc.')
        self.ctl.output('local_status <appli>:*\t\t\t'
                        'Get the local status of all processes in the application named appli.')
        self.ctl.output('local_status <proc> <proc>\t\tGet the local status for multiple named processes')
        self.ctl.output('local_status\t\t\t\tGet the local status of all processes.')

    def do_process_rules(self, arg):
        """ Command to get the process rules handled by Supvisors. """
        if self._upcheck():
            namespecs = arg.split()
            if not namespecs or 'all' in namespecs:
                try:
                    namespecs = [make_namespec(info['application_name'], info['process_name'])
                                 for info in self.supvisors().get_all_process_info()]
                except xmlrpclib.Fault as e:
                    self.ctl.output(f'ERROR ({e.faultString})')
                    self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                    namespecs = []
            rules_list = []
            for namespec in sorted(namespecs):
                try:
                    rules = self.supvisors().get_process_rules(namespec)
                except xmlrpclib.Fault as e:
                    self.ctl.output(f'{namespec}: ERROR ({e.faultString})')
                    self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                else:
                    rules_list.extend(rules)
            # print results
            if rules_list:
                max_appli = ControllerPlugin.max_template(rules_list, 'application_name', 'Application')
                max_proc = ControllerPlugin.max_template(rules_list, 'process_name', 'Process')
                template = (f'%(appli)-{max_appli}s%(proc)-{max_proc}s%(start_seq)-7s%(stop_seq)-7s'
                            '%(req)-12s%(exit)-12s%(load)-12s%(strategy)-22s%(identifiers)s')
                # print title
                payload = {'appli': 'Application', 'proc': 'Process', 'start_seq': 'Start', 'stop_seq': 'Stop',
                           'req': 'Required', 'exit': 'WaitExit', 'load': 'Loading', 'strategy': 'Strategy',
                           'identifiers': 'Supervisor'}
                self.ctl.output(template % payload)
                # print filtered payloads
                for rules in rules_list:
                    payload = {'appli': rules['application_name'],
                               'proc': rules['process_name'],
                               'identifiers': rules['identifiers'],
                               'start_seq': rules['start_sequence'],
                               'stop_seq': rules['stop_sequence'],
                               'req': rules['required'],
                               'exit': rules['wait_exit'],
                               'load': f"{rules['expected_loading']}%",
                               'strategy': rules['running_failure_strategy']}
                    self.ctl.output(template % payload)

    def help_process_rules(self):
        """ Print the help of the process rules command."""
        self.ctl.output('process_rules <proc>\t\t\tGet the rules of the process named proc.')
        self.ctl.output('process_rules <appli>:*\t\t\tGet the rules of all processes in the application named appli.')
        self.ctl.output('process_rules <proc> <proc>\t\tGet the rules for multiple named processes')
        self.ctl.output('process_rules\t\t\t\tGet the rules of all processes.')

    def do_conflicts(self, _):
        """ Command to get the conflicts detected by Supvisors. """
        if self._upcheck():
            try:
                conflicts = self.supvisors().get_conflicts()
            except xmlrpclib.Fault as e:
                self.ctl.output(f'ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
            else:
                if conflicts:
                    max_appli = ControllerPlugin.max_template(conflicts, 'application_name', 'Application')
                    max_proc = ControllerPlugin.max_template(conflicts, 'process_name', 'Process')
                    template = f'%(appli)-{max_appli}s%(proc)-{max_proc}s%(state)-12s%(identifiers)s'
                    # print title
                    payload = {'appli': 'Application', 'proc': 'Process', 'state': 'State', 'identifiers': 'Supervisor'}
                    self.ctl.output(template % payload)
                    # print filtered payloads
                    for conflict in conflicts:
                        payload = {'appli': conflict['application_name'],
                                   'proc': conflict['process_name'],
                                   'state': conflict['statename'],
                                   'identifiers': conflict['identifiers']}
                        self.ctl.output(template % payload)

    def help_conflicts(self):
        """ Print the help of the conflicts command."""
        self.ctl.output('conflicts\t\t\t\tGet the Supvisors conflicts.')

    def _get_matching_applications(self, parameters, all_info, alt_rpc):
        """ Get the matching information from the full list. """
        matching_info = []
        if not parameters or 'all' in parameters:
            # get all managed applications
            matching_info = sorted([application_name for application_name, managed in all_info.items() if managed])
        else:
            for application_name in parameters:
                if application_name in all_info:
                    if all_info[application_name]:
                        matching_info.append(application_name)
                    else:
                        self.ctl.output(f'{application_name}: IGNORED (unmanaged. use {alt_rpc} {application_name}:*)')
                else:
                    self.ctl.output(f'{application_name}: ERROR (no such application)')
                    self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
        return matching_info

    def do_start_application(self, arg):
        """ Command to start Supvisors applications using a strategy and rules. """
        if self._upcheck():
            args = as_string(arg).split()
            # check number of arguments
            if len(args) < 1:
                self.ctl.output('ERROR: start_application requires at least a strategy')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_start_application()
                return
            # check strategy format
            try:
                strategy = StartingStrategies[args[0]]
            except KeyError:
                self.ctl.output('ERROR: unknown strategy for start_application.'
                                f' use one of {[x.name for x in StartingStrategies]}')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_start_application()
                return
            # get all application info
            try:
                all_info = {application_info['application_name']: application_info['managed']
                            for application_info in self.supvisors().get_all_applications_info()}
            except xmlrpclib.Fault as e:
                self.ctl.output(f'ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                return
            # match with parameters
            matching_info = self._get_matching_applications(args[1:], all_info, 'start')
            # request start for matching applications
            for application_name in matching_info:
                try:
                    result = self.supvisors().start_application(strategy.value, application_name)
                except xmlrpclib.Fault as e:
                    self.ctl.output(f'{application_name}: ERROR ({e.faultString})')
                    self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                else:
                    self.ctl.output(f'{application_name} started: {result}')

    def help_start_application(self):
        """ Print the help of the start_application command."""
        self.ctl.output('start_application <strategy> <appli>\t\t'
                        'Start the managed application named appli with strategy.')
        self.ctl.output('start_application <strategy> <appli> <appli>\t'
                        'Start multiple managed named applications with strategy')
        self.ctl.output('start_application <strategy>\t\t\t'
                        'Start all managed applications with strategy.')

    def do_test_start_application(self, arg):
        """ Command to test the starting of Supvisors applications using a strategy and rules. """
        if self._upcheck():
            args = as_string(arg).split()
            # check number of arguments
            if len(args) != 2:
                self.ctl.output('ERROR: test_start_application requires one strategy and one application name')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_test_start_application()
                return
            # check strategy format
            try:
                strategy = StartingStrategies[args[0]]
            except KeyError:
                self.ctl.output('ERROR: unknown strategy for test_start_application.'
                                f' use one of {[x.name for x in StartingStrategies]}')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_test_start_application()
                return
            # get all application info
            application_name = args[1]
            try:
                info = self.supvisors().get_application_info(args[1])
            except xmlrpclib.Fault:
                self.ctl.output(f'ERROR: unknown application "{application_name}"')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                return
            # match with parameters
            if not info['managed']:
                self.ctl.output(f'ERROR: {application_name} unmanaged')
                self.help_test_start_application()
                return
            # request test start for matching applications
            try:
                results_list = self.supvisors().test_start_application(strategy.value, application_name)
            except xmlrpclib.Fault as e:
                self.ctl.output(f'{application_name}: ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                return
            # print results summary
            max_appli = ControllerPlugin.max_template(results_list, 'application_name', 'Application')
            max_process = ControllerPlugin.max_template(results_list, 'process_name', 'Process')
            max_identifiers = ControllerPlugin.max_template(results_list, 'running_identifiers', 'Supervisor')
            # print title
            template = (f'%(appli)-{max_appli}s%(proc)-{max_process}s%(state)-12s'
                        f'%(identifiers)-{max_identifiers}s%(reason)s')
            title = {'appli': 'Application', 'proc': 'Process', 'state': 'State',
                     'identifiers': 'Supervisor', 'reason': 'Reason'}
            self.ctl.output(template % title)
            # print results
            for results in results_list:
                payload = {'appli': results['application_name'],
                           'proc': results['process_name'],
                           'identifiers': results['running_identifiers'],
                           'state': results['state'],
                           'reason': results['forced_reason']}
                self.ctl.output(template % payload)

    def help_test_start_application(self):
        """ Print the help of the test_start_application command."""
        self.ctl.output('test_start_application <strategy> <appli>\t\t'
                        'Test the starting of the managed application named appli with strategy.')
        self.ctl.output('test_start_application <strategy>\t\t\t'
                        'Test the starting of all managed applications with strategy.')

    def do_restart_application(self, arg):
        """ Command to restart Supvisors applications using a strategy and rules. """
        if self._upcheck():
            args = as_string(arg).split()
            # check number of arguments
            if len(args) < 1:
                self.ctl.output('ERROR: restart_application requires at least a strategy')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_restart_application()
                return
            # check strategy format
            try:
                strategy = StartingStrategies[args[0]]
            except KeyError:
                self.ctl.output('ERROR: unknown strategy for restart_application.'
                                f' use one of {[x.name for x in StartingStrategies]}')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_restart_application()
                return
            # get all application info
            try:
                all_info = {application_info['application_name']: application_info['managed']
                            for application_info in self.supvisors().get_all_applications_info()}
            except xmlrpclib.Fault as e:
                self.ctl.output(f'ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                return
            # match with parameters
            matching_info = self._get_matching_applications(args[1:], all_info, 'restart')
            # request start for matching applications
            for application_name in matching_info:
                try:
                    result = self.supvisors().restart_application(strategy.value, application_name)
                except xmlrpclib.Fault as e:
                    self.ctl.output(f'{application_name}: ERROR ({e.faultString})')
                    self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                else:
                    self.ctl.output(f'{application_name} restarted: {result}')

    def help_restart_application(self):
        """ Print the help of the restart_application command."""
        self.ctl.output('restart_application <strategy> <appli>\t\t'
                        'Restart the managed application named appli with strategy.')
        self.ctl.output('restart_application <strategy> <appli> <appli>\t'
                        'Restart multiple managed named applications with strategy')
        self.ctl.output('restart_application <strategy> \t\t\t'
                        'Restart all managed applications with strategy.')

    def do_stop_application(self, arg):
        """ Command to stop Supvisors applications using rules. """
        if self._upcheck():
            # get all application info
            try:
                all_info = {application_info['application_name']: application_info['managed']
                            for application_info in self.supvisors().get_all_applications_info()}
            except xmlrpclib.Fault as e:
                self.ctl.output('ERROR ({})'.format(e.faultString))
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                return
            # match with parameters
            args = as_string(arg).split()
            matching_info = self._get_matching_applications(args, all_info, 'stop')
            # request start for matching applications
            for application_name in matching_info:
                try:
                    result = self.supvisors().stop_application(application_name)
                except xmlrpclib.Fault as e:
                    self.ctl.output('{}: ERROR ({})'.format(application_name, e.faultString))
                    self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                else:
                    self.ctl.output('{} stopped: {}'.format(application_name, result))

    def help_stop_application(self):
        """ Print the help of the stop_application command."""
        self.ctl.output('stop_application <appli>\t\t\tStop the managed application named appli.')
        self.ctl.output('stop_application <appli> <appli>\t\tStop multiple managed named applications.')
        self.ctl.output('stop_application\t\t\t\tStop all managed applications.')

    def do_all_start(self, arg):
        """ Invoke the same Supervisor start command over all running nodes. """
        if self._upcheck():
            args = arg.split()
            if len(args) < 1:
                self.ctl.output('ERROR: all_start requires a process name')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_all_start()
                return
            namespec = args[0]
            # get running instances
            urls = self.get_running_instances()
            # send the start command to all instances
            if urls:
                for identifier, url in urls.items():
                    try:
                        # no wait
                        proxy = self.get_server_proxy(url, 'supervisor')
                        result = proxy.startProcess(namespec, False)
                    except xmlrpclib.Fault as e:
                        self.ctl.output(f'{namespec}: ERROR ({e.faultString})')
                        self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                    else:
                        self.ctl.output(f'{namespec} started on {identifier}: {result}')
            else:
                self.ctl.output(f'ERROR: no running Supvisors instance')

    def help_all_start(self):
        """ Print the help of the all_start command."""
        self.ctl.output('all_start <proc>\t\t\t'
                        'Start the process named proc on all RUNNING Supvisors instances.')

    def do_start_args(self, arg):
        """ Command to start a local process with additional arguments. """
        if self._upcheck():
            args = arg.split()
            if len(args) < 2:
                self.ctl.output('ERROR: start_args requires a process name and extra arguments')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_start_args()
                return
            namespec = args[0]
            try:
                result = self.supvisors().start_args(namespec, ' '.join(args[1:]))
            except xmlrpclib.Fault as e:
                self.ctl.output(f'{namespec}: ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
            else:
                self.ctl.output(f'{namespec} started: {result}')

    def help_start_args(self):
        """ Print the help of the start_args command."""
        self.ctl.output('start_args <proc> <arg_list>\t\t\t'
                        'Start the local process named proc with additional arguments arg_list.')

    def do_all_start_args(self, arg):
        """ Invoke the same Supvisors start_args command over all running nodes. """
        if self._upcheck():
            args = arg.split()
            if len(args) < 2:
                self.ctl.output('ERROR: all_start_args requires a process name and extra arguments')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_all_start_args()
                return
            namespec = args[0]
            # get running instances
            urls = self.get_running_instances()
            if urls:
                # send the start command to all instances
                for identifier, url in urls.items():
                    try:
                        # no wait
                        result = self.get_server_proxy(url, 'supvisors').start_args(namespec, ' '.join(args[1:]), False)
                    except xmlrpclib.Fault as e:
                        self.ctl.output(f'{namespec}: ERROR ({e.faultString})')
                        self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                    else:
                        self.ctl.output(f'{namespec} started on {identifier}: {result}')
            else:
                self.ctl.output(f'ERROR: no running Supvisors instance')

    def help_all_start_args(self):
        """ Print the help of the all_start_args command."""
        self.ctl.output('all_start_args <proc> <arg_list>\t\t'
                        'Start the process named proc on all RUNNING Supvisors instances'
                        ' with additional arguments arg_list.')

    def do_start_process(self, arg):
        """ Command to start processes with a strategy and rules. """
        if self._upcheck():
            args = arg.split()
            if len(args) < 1:
                self.ctl.output('ERROR: start_process requires at least a strategy')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_start_process()
                return
            try:
                strategy = StartingStrategies[args[0]]
            except KeyError:
                self.ctl.output('ERROR: unknown strategy for start_process.'
                                f' use one of {[x.name for x in StartingStrategies]}')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_start_process()
                return
            namespecs = args[1:]
            if not namespecs or 'all' in namespecs:
                try:
                    namespecs = [make_namespec(info['application_name'], info['process_name'])
                                 for info in self.supvisors().get_all_process_info()]
                except xmlrpclib.Fault as e:
                    self.ctl.output(f'ERROR ({e.faultString})')
                    self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                    namespecs = []
            for namespec in sorted(namespecs):
                try:
                    result = self.supvisors().start_process(strategy.value, namespec)
                except xmlrpclib.Fault as e:
                    self.ctl.output(f'{namespec}: ERROR ({e.faultString})')
                    self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                else:
                    self.ctl.output(f'{namespec} started: {result}')

    def help_start_process(self):
        """ Print the help of the start_process command."""
        self.ctl.output('start_process <strategy> <proc>\t\t\tStart the process named proc with strategy.')
        self.ctl.output('start_process <strategy> <proc> <proc>\t\tStart multiple named processes with strategy.')
        self.ctl.output('start_process <strategy>\t\t\tStart all processes with strategy.')

    def do_test_start_process(self, arg):
        """ Command to test the starting of Supvisors processes using a strategy and rules. """
        if self._upcheck():
            args = arg.split()
            if len(args) != 2:
                self.ctl.output('ERROR: test_start_process requires one strategy and one namespec')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_test_start_process()
                return
            try:
                strategy = StartingStrategies[args[0]]
            except KeyError:
                self.ctl.output('ERROR: unknown strategy for test_start_process.'
                                f' use one of {[x.name for x in StartingStrategies]}')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_test_start_process()
                return
            namespec = args[1]
            try:
                results_list = self.supvisors().test_start_process(strategy.value, namespec)
            except xmlrpclib.Fault as e:
                self.ctl.output(f'{namespec}: ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                return
            # print results summary
            max_appli = ControllerPlugin.max_template(results_list, 'application_name', 'Application')
            max_process = ControllerPlugin.max_template(results_list, 'process_name', 'Process')
            max_identifiers = ControllerPlugin.max_template(results_list, 'running_identifiers', 'Supervisor')
            # print title
            template = (f'%(appli)-{max_appli}s%(proc)-{max_process}s%(state)-12s'
                        f'%(identifiers)-{max_identifiers}s%(reason)s')
            title = {'appli': 'Application', 'proc': 'Process', 'state': 'State',
                     'identifiers': 'Supervisor', 'reason': 'Reason'}
            self.ctl.output(template % title)
            # print results
            for results in results_list:
                payload = {'appli': results['application_name'],
                           'proc': results['process_name'],
                           'identifiers': results['running_identifiers'],
                           'state': results['state'],
                           'reason': results['forced_reason']}
                self.ctl.output(template % payload)

    def help_test_start_process(self):
        """ Print the help of the test_start_process command."""
        self.ctl.output('test_start_process <strategy> <proc>\t\t\tStart the process named proc with strategy.')
        self.ctl.output('test_start_process <strategy>\t\t\tStart all processes with strategy.')

    def do_start_any_process(self, arg):
        """ Command to start processes using regular expressions, with a strategy and rules. """
        if self._upcheck():
            args = arg.split()
            if len(args) < 2:
                self.ctl.output('ERROR: start_any_process requires a strategy and at least a regular expression')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_start_any_process()
                return
            try:
                strategy = StartingStrategies[args[0]]
            except KeyError:
                self.ctl.output('ERROR: unknown strategy for start_any_process.'
                                f' use one of {[x.name for x in StartingStrategies]}')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_start_any_process()
                return
            regexes = args[1:]
            for regex in sorted(regexes):
                try:
                    result = self.supvisors().start_any_process(strategy.value, regex)
                except xmlrpclib.Fault as e:
                    self.ctl.output(f'"{regex}": ERROR ({e.faultString})')
                    self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                else:
                    self.ctl.output(f'{result} started')

    def help_start_any_process(self):
        """ Print the help of the start_any_process command."""
        self.ctl.output('start_any_process <strategy> <regex>\t\t\t'
                        'Start a process whose namespec matches the regular expression and with a starting strategy.')
        self.ctl.output('start_any_process <strategy> <regex> <regex>\t\t'
                        'Start multiple processes whose namespec matches the regular expressions'
                        ' and with a starting strategy.')

    # start a process using strategy, rules and additional arguments
    def do_start_process_args(self, arg):
        """ Command to start a process with a strategy, rules and additional arguments. """
        if self._upcheck():
            args = arg.split()
            if len(args) < 3:
                self.ctl.output('ERROR: start_process_args requires a strategy, '
                                'a program namespec and extra arguments')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_start_process_args()
                return
            try:
                strategy = StartingStrategies[args[0]]
            except KeyError:
                self.ctl.output('ERROR: unknown strategy for start_process_args.'
                                f' use one of {[x.name for x in StartingStrategies]}')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_start_process_args()
                return
            namespec = args[1]
            try:
                result = self.supvisors().start_process(strategy.value, namespec, ' '.join(args[2:]))
            except xmlrpclib.Fault as e:
                self.ctl.output(f'{namespec}: ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
            else:
                self.ctl.output(f'{namespec} started: {result}')

    def help_start_process_args(self):
        """ Print the help of the start_process_args command."""
        self.ctl.output('start_process <strategy> <proc> <arg_list>\t'
                        'Start the process named proc with additional arguments arg_list.')

    def do_start_any_process_args(self, arg):
        """ Command to start processes using regular expressions, with a strategy and rules. """
        if self._upcheck():
            args = arg.split()
            if len(args) < 3:
                self.ctl.output('ERROR: start_any_process_args requires a strategy, a regular expression'
                                ' and extra arguments')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_start_any_process_args()
                return
            try:
                strategy = StartingStrategies[args[0]]
            except KeyError:
                self.ctl.output('ERROR: unknown strategy for start_any_process_args.'
                                f' use one of {[x.name for x in StartingStrategies]}')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_start_any_process_args()
                return
            regex = args[1]
            try:
                result = self.supvisors().start_any_process(strategy.value, regex, ' '.join(args[2:]))
            except xmlrpclib.Fault as e:
                self.ctl.output(f'"{regex}": ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
            else:
                self.ctl.output(f'{result} started')

    def help_start_any_process_args(self):
        """ Print the help of the start_any_process_args command."""
        self.ctl.output('start_any_process_args <strategy> <regex> <arg_list>\t'
                        'Start a process whose namespec matches the regular expression, using a starting strategy and'
                        ' additional arguments arg_list.')

    def do_stop_process(self, arg):
        """ Command to stop processes with rules. """
        if self._upcheck():
            namespecs = arg.split()
            if not namespecs or 'all' in namespecs:
                try:
                    namespecs = [make_namespec(info['application_name'], info['process_name'])
                                 for info in self.supvisors().get_all_process_info()]
                except xmlrpclib.Fault as e:
                    self.ctl.output(f'ERROR ({e.faultString})')
                    self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                    namespecs = []
            for namespec in sorted(namespecs):
                try:
                    self.supvisors().stop_process(namespec)
                except xmlrpclib.Fault as e:
                    self.ctl.output(f'{namespec}: ERROR ({e.faultString})')
                    self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                else:
                    self.ctl.output(f'{namespec} stopped')

    def help_stop_process(self):
        """ Print the help of the stop_process command."""
        self.ctl.output('stop_process <proc>\t\tStop the process named proc.')
        self.ctl.output('stop_process <proc> <proc>\tStop multiple named processes.')
        self.ctl.output('stop_process\t\t\tStop all named processes.')

    def do_restart_process(self, arg):
        """ Command to restart processes with a strategy and rules. """
        if self._upcheck():
            args = arg.split()
            if len(args) < 1:
                self.ctl.output('ERROR: restart_process requires a strategy and a program name')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_restart_process()
                return
            try:
                strategy = StartingStrategies[args[0]]
            except KeyError:
                self.ctl.output(f'ERROR: unknown strategy={args[0]}. use one of {[x.name for x in StartingStrategies]}')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_restart_process()
                return
            processes = args[1:]
            if not processes or 'all' in processes:
                try:
                    processes = [make_namespec(info['application_name'], info['process_name'])
                                 for info in self.supvisors().get_all_process_info()]
                except xmlrpclib.Fault as e:
                    self.ctl.output('ERROR ({})'.format(e.faultString))
                    self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                    processes = []
            for process in sorted(processes):
                try:
                    result = self.supvisors().restart_process(strategy.value, process)
                except xmlrpclib.Fault as e:
                    self.ctl.output('{}: ERROR ({})'.format(process, e.faultString))
                    self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                else:
                    self.ctl.output('{} restarted: {}'.format(process, result))

    def help_restart_process(self):
        """ Print the help of the restart_process command."""
        self.ctl.output("restart_process <strategy> <proc>\t\t"
                        "Restart the process named proc using strategy.")
        self.ctl.output("restart_process <strategy> <proc> <proc>\t"
                        "Restart multiple named processes using strategy.")
        self.ctl.output("restart_process <strategy> \t\t\t"
                        "Restart all processes using strategy.")

    def do_update_numprocs(self, arg):
        """ Command to dynamically update the numprocs of the program.
        Implementation of Supervisor issue #177 - Dynamic numproc change.
        """
        if self._upcheck():
            args = arg.split()
            if len(args) < 2:
                self.ctl.output('ERROR: update_numprocs requires a program name and a numprocs values')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_update_numprocs()
                return
            try:
                value = int(args[1])
                assert value > 0
            except (ValueError, AssertionError):
                self.ctl.output('ERROR: numprocs must be a strictly positive integer')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_update_numprocs()
                return
            try:
                result = self.supvisors().update_numprocs(args[0], value)
            except xmlrpclib.Fault as e:
                self.ctl.output(f'ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
            else:
                self.ctl.output(f'{args[0]} numprocs updated: {result}')

    def help_update_numprocs(self):
        """ Print the help of the update_numprocs command. """
        self.ctl.output('update_numprocs program_name numprocs\t\tUpdate the program numprocs.')

    def do_lazy_update_numprocs(self, arg):
        """ Command to dynamically update the numprocs of the program.
        Implementation of Supervisor issue #177 - Dynamic numproc change.
        The difference with update_numprocs is that the obsolete processes will be removed from the configuration
        only when they will stop (by themselves or as a consequence of a later action).
        """
        if self._upcheck():
            args = arg.split()
            if len(args) < 2:
                self.ctl.output('ERROR: lazy_update_numprocs requires a program name and a numprocs values')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_lazy_update_numprocs()
                return
            try:
                value = int(args[1])
                assert value > 0
            except (ValueError, AssertionError):
                self.ctl.output('ERROR: numprocs must be a strictly positive integer')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_lazy_update_numprocs()
                return
            try:
                result = self.supvisors().update_numprocs(args[0], value, True, True)
            except xmlrpclib.Fault as e:
                self.ctl.output(f'ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
            else:
                self.ctl.output(f'{args[0]} numprocs updated: {result}')

    def help_lazy_update_numprocs(self):
        """ Print the help of the lazy_update_numprocs command. """
        self.ctl.output('lazy_update_numprocs program_name numprocs\t\t\t\tUpdate the program numprocs (lazy mode).')

    def do_enable(self, arg):
        """ Command to enable the processes corresponding to the program.
        Implementation of Supervisor issue #591 - New Feature: disable/enable.
        """
        if self._upcheck():
            args = arg.split()
            if len(args) < 1:
                self.ctl.output('ERROR: enable requires a program name')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_enable()
                return
            try:
                result = self.supvisors().enable(args[0])
            except xmlrpclib.Fault as e:
                self.ctl.output(f'ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
            else:
                self.ctl.output(f'{args[0]} enabled: {result}')

    def help_enable(self):
        """ Print the help of the enable command. """
        self.ctl.output('enable program_name\t\t\tEnable the program.')

    def do_disable(self, arg):
        """ Command to disable the processes corresponding to the program.
        Implementation of Supervisor issue #591 - New Feature: disable/enable.
        """
        if self._upcheck():
            args = arg.split()
            if len(args) < 1:
                self.ctl.output('ERROR: disable requires a program name')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_disable()
                return
            try:
                result = self.supvisors().disable(args[0])
            except xmlrpclib.Fault as e:
                self.ctl.output(f'ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
            else:
                self.ctl.output(f'{args[0]} disabled: {result}')

    def help_disable(self):
        """ Print the help of the disable command. """
        self.ctl.output('disable program_name\t\t\tDisable the program.')

    def do_conciliate(self, arg):
        """ Command to conciliate conflicts (applicable with default USER strategy). """
        if self._upcheck():
            args = arg.split()
            if len(args) < 1:
                self.ctl.output('ERROR: conciliate requires a strategy')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_conciliate()
                return
            try:
                strategy = ConciliationStrategies[args[0]]
            except KeyError:
                self.ctl.output(f'ERROR: unknown strategy for conciliate.'
                                f' use one of {[x.name for x in ConciliationStrategies]}')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_conciliate()
                return
            try:
                result = self.supvisors().conciliate(strategy.value)
            except xmlrpclib.Fault as e:
                self.ctl.output(f'ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
            else:
                self.ctl.output(f'Conciliated: {result}')

    def help_conciliate(self):
        """ Print the help of the conciliate command. """
        self.ctl.output('conciliate <strategy>\t\t\tConciliate process conflicts using strategy')

    def do_restart_sequence(self, _):
        """ Command to trigger the whole start sequence of Supvisors. """
        if self._upcheck():
            try:
                result = self.supvisors().restart_sequence()
            except xmlrpclib.Fault as e:
                self.ctl.output(f'ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
            else:
                self.ctl.output(f'Start sequence completed: {result}')

    def help_restart_sequence(self):
        """ Print the help of the restart_sequence command."""
        self.ctl.output('restart_sequence\t\t\t\t\tTrigger the whole start sequence')

    def do_sreload(self, _):
        """ Command to restart Supvisors on all instances. """
        if self._upcheck():
            try:
                result = self.supvisors().restart()
            except xmlrpclib.Fault as e:
                self.ctl.output(f'ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
            else:
                self.ctl.output(f'Restarted: {result}')

    def help_sreload(self):
        """ Print the help of the sreload command."""
        self.ctl.output('sreload\t\t\t\t\tRestart Supvisors on all instances')

    def do_sshutdown(self, _):
        """ Command to shut down Supvisors on all instances. """
        if self._upcheck():
            try:
                result = self.supvisors().shutdown()
            except xmlrpclib.Fault as e:
                self.ctl.output(f'ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
            else:
                self.ctl.output(f'Shut down: {result}')

    def help_sshutdown(self):
        """ Print the help of the sshutdown command."""
        self.ctl.output('sshutdown\t\t\t\tShutdown Supvisors on all instances')

    def do_end_sync(self, arg):
        """ Command to end the Supvisors synchronization phase. """
        if self._upcheck():
            # check request args
            args = arg.split()
            if len(args) > 1:
                self.ctl.output('ERROR: end_sync optionally takes one single Supvisors identifier')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_end_sync()
                return
            master = args[0] if args else ''
            try:
                result = self.supvisors().end_sync(master)
            except xmlrpclib.Fault as e:
                self.ctl.output(f'ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
            else:
                self.ctl.output(f'End of Synchronization phase requested: {result}')

    def help_end_sync(self):
        """ Print the help of the end_sync command."""
        self.ctl.output('end_sync\t\t\t\tEnd the Supvisors synchronization phase')
        self.ctl.output('end_sync master_identifier\t\tEnd the Supvisors synchronization phase'
                        ' and select this identifier as Supvisors Master instance ')

    def do_loglevel(self, arg):
        """ Command to change the level of the local Supvisors. """
        if self._upcheck():
            args = arg.split()
            if len(args) < 1:
                self.ctl.output('ERROR: loglevel requires a level')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_loglevel()
                return
            level = getLevelNumByDescription(args[0])
            if level is None:
                self.ctl.output('ERROR: unknown level for Logger')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_loglevel()
                return
            try:
                result = self.supvisors().change_log_level(level)
            except xmlrpclib.Fault as e:
                self.ctl.output(f'ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
            else:
                self.ctl.output(f'Logger level changed: {result}')

    def help_loglevel(self):
        """ Print the help of the loglevel command."""
        self.ctl.output('loglevel lvl\t\t\t\tChange the level of local Supvisors\' logger to lvl.')
        values = RPCInterface.get_logger_levels().values()
        self.ctl.output(f'\t\t\t\t\tApplicable values are: {values}.')

    def do_enable_stats(self, arg):
        """ Command to enable host/process statistics collection in the local Supvisors. """
        if self._upcheck():
            args = arg.split()
            enable_host_stats, enable_proc_stats = True, True
            if args:
                enable_host_stats, enable_proc_stats = False, False
                error = False
                # check request args
                stats_types = list(map(str.lower, arg.split()))
                for stats_type in stats_types:
                    if stats_type == 'host':
                        enable_host_stats = True
                    elif stats_type == 'process':
                        enable_proc_stats = True
                    elif stats_type == 'all':
                        enable_host_stats = True
                        enable_proc_stats = True
                    else:
                        self.ctl.output(f'ERROR: invalid statistics type {stats_type}')
                        error = True
                if error:
                    self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                    self.help_enable_stats()
                    return
            # disable host statistics collection
            if enable_host_stats:
                try:
                    result = self.supvisors().enable_host_statistics(True)
                except xmlrpclib.Fault as e:
                    self.ctl.output(f'ERROR ({e.faultString})')
                    self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                else:
                    self.ctl.output(f'Host statistics enabled: {result}')
            # disable process statistics collection
            if enable_proc_stats:
                try:
                    result = self.supvisors().enable_process_statistics(True)
                except xmlrpclib.Fault as e:
                    self.ctl.output(f'ERROR ({e.faultString})')
                    self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                else:
                    self.ctl.output(f'Process statistics enabled: {result}')

    def help_enable_stats(self):
        """ Print the help of the enable_stats command."""
        self.ctl.output('enable_stats\t\t\tEnable host and process statistics collection')
        self.ctl.output('enable_stats all\t\tEnable host and process statistics collection')
        self.ctl.output('enable_stats host\t\tEnable host statistics collection')
        self.ctl.output('enable_stats process\t\tEnable process statistics collection')
        self.ctl.output('enable_stats host process\tEnable host and process statistics collection')

    def do_disable_stats(self, arg):
        """ Command to disable host/process statistics collection in the local Supvisors. """
        if self._upcheck():
            args = arg.split()
            disable_host_stats, disable_proc_stats = True, True
            if args:
                disable_host_stats, disable_proc_stats = False, False
                error = False
                # check request args
                stats_types = list(map(str.lower, arg.split()))
                for stats_type in stats_types:
                    if stats_type == 'host':
                        disable_host_stats = True
                    elif stats_type == 'process':
                        disable_proc_stats = True
                    elif stats_type == 'all':
                        disable_host_stats = True
                        disable_proc_stats = True
                    else:
                        self.ctl.output(f'ERROR: invalid statistics type {stats_type}')
                        error = True
                if error:
                    self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                    self.help_disable_stats()
                    return
            # disable host statistics collection
            if disable_host_stats:
                try:
                    result = self.supvisors().enable_host_statistics(False)
                except xmlrpclib.Fault as e:
                    self.ctl.output(f'ERROR ({e.faultString})')
                    self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                else:
                    self.ctl.output(f'Host statistics disabled: {result}')
            # disable process statistics collection
            if disable_proc_stats:
                try:
                    result = self.supvisors().enable_process_statistics(False)
                except xmlrpclib.Fault as e:
                    self.ctl.output(f'ERROR ({e.faultString})')
                    self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
                else:
                    self.ctl.output(f'Process statistics disabled: {result}')

    def help_disable_stats(self):
        """ Print the help of the disable_stats command."""
        self.ctl.output('disable_stats\t\t\tDisable host and process statistics collection')
        self.ctl.output('disable_stats all\t\tDisable host and process statistics collection')
        self.ctl.output('disable_stats host\t\tDisable host statistics collection')
        self.ctl.output('disable_stats process\t\tDisable process statistics collection')
        self.ctl.output('disable_stats host process\tDisable host and process statistics collection')

    def do_stats_period(self, arg):
        """ Command to enable/disable host/process statistics collection in the local Supvisors. """
        if self._upcheck():
            # check request args
            args = arg.split()
            if len(args) < 1:
                self.ctl.output('ERROR: stats_period requires a period')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_stats_period()
                return
            try:
                value = float(args[0])
                assert value > 0.0
            except ValueError:
                self.ctl.output('ERROR: stats_period requires a strictly positive period')
                self.ctl.exitstatus = LSBInitExitStatuses.INVALID_ARGS
                self.help_stats_period()
                return
            try:
                result = self.supvisors().update_collecting_period(value)
            except xmlrpclib.Fault as e:
                self.ctl.output(f'ERROR ({e.faultString})')
                self.ctl.exitstatus = LSBInitExitStatuses.GENERIC
            else:
                self.ctl.output(f'Statistics period updated: {result}')

    def help_stats_period(self):
        """ Print the help of the stats_period command."""
        self.ctl.output('stats_period period\t\tUpdate the host and process statistics collection period')

    def _upcheck(self):
        """ Check of the API versions. """
        try:
            api = self.supvisors().get_api_version()
            if api != __version__:
                self.ctl.output('ERROR: this version of supvisorsctl expects to talk to a server '
                                'with API version %s, but the remote version is %s.' % (__version__, api))
                self.ctl.exitstatus = LSBInitExitStatuses.NOT_INSTALLED
                return False
        except xmlrpclib.Fault as e:
            if e.faultCode == xmlrpc.Faults.UNKNOWN_METHOD:
                self.ctl.output('ERROR: supervisord responded but did not recognize '
                                'the supvisors namespace commands that supervisorstl uses to control it. '
                                'Please check that the [rpcinterface:supvisors] section is enabled '
                                'in the configuration file.')
                self.ctl.exitstatus = LSBInitExitStatuses.UNIMPLEMENTED_FEATURE
                return False
            self.exitstatus = LSBInitExitStatuses.GENERIC
            raise
        except socket.error as why:
            if why.args[0] == errno.ECONNREFUSED:
                self.ctl.output(f'ERROR: {self.ctl.options.serverurl} refused connection')
                self.ctl.exitstatus = LSBInitExitStatuses.INSUFFICIENT_PRIVILEGES
                return False
            elif why.args[0] == errno.ENOENT:
                self.ctl.output(f'ERROR: {self.ctl.options.serverurl} no such file')
                self.ctl.exitstatus = LSBInitExitStatuses.NOT_RUNNING
                return False
            self.exitstatus = LSBInitExitStatuses.GENERIC
            raise
        return True


def make_supvisors_controller_plugin(controller):
    """ Create a plugin for the Supvisors commands. """
    return ControllerPlugin(controller)


# Copied and adapted from supervisor.supervisorctl source code
def main(args=None, options=None):
    # read options
    if options is None:
        options = ClientOptions()
    options.realize(args, doc=supervisorctl.__doc__)
    # add supvisors plugin if not there
    if not any(factory[0] == 'supvisors' for factory in options.plugin_factories):
        options.plugin_factories.append(('supvisors', make_supvisors_controller_plugin, {}))
    # create controller
    c = Controller(options)
    # process single command
    if options.args:
        c.onecmd(' '.join(options.args))
        sys.exit(c.exitstatus)
    # enter the interactive mode
    if options.interactive:
        c.exec_cmdloop(args, options)
        sys.exit(0)  # exitstatus always 0 for interactive mode



# ======================================================================
# Copyright 2016 Julien LE CLEACH
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import socket
from unittest.mock import call

import pytest

from supvisors.internal_com.mapper import *


def test_get_addresses(supvisors):
    """ Test the SupvisorsMapper.get_addresses method. """
    # complex to test as it depends on the network configuration of the operating system
    # check that there is at least one entry looking like an IPv4 address
    host_name, aliases, ip_list = get_addresses(gethostname(), supvisors.logger)
    assert re.match(r'^\d{1,3}(.\d{1,3}){3}$', ip_list[0])
    assert host_name == getfqdn()
    # test exception on unknown IP address and unknown host name
    assert get_addresses('10.4', supvisors.logger) is None
    assert get_addresses('unknown node', supvisors.logger) is None


def test_sup_id_create_no_match(supvisors):
    """ Test the values set at SupvisorsInstanceId construction. """
    no_matches = ['', 'ident>', 'cliche81:12000:', '10.0.0.1:145000']
    # test with no match but default options loaded
    for item in no_matches:
        sup_id = SupvisorsInstanceId(item, supvisors)
        assert sup_id.identifier is None
        assert sup_id.nick_identifier is None
        assert sup_id.host_id is None
        assert sup_id.http_port == 25000  # defaulted to supervisor server port
        assert sup_id.event_port == 25200  # defaulted to options.event_port
        assert sup_id.host_name is None
        assert sup_id.aliases is None
        assert sup_id.ip_addresses is None
        assert sup_id.ip_address is None
        assert sup_id.stereotypes == []
        with pytest.raises(TypeError):
            str(sup_id)
        assert sup_id.serial() == {'host_id': None, 'host_name': None, 'http_port': 25000,
                                   'identifier': None, 'ip_addresses': None,
                                   'nick_identifier': None, 'stereotypes': []}


def test_sup_id_create_simple_no_default(supvisors):
    """ Test the values set at SupvisorsInstanceId construction. """
    supvisors.options.event_port = 0
    # test with no match and no default option loaded
    sup_id = SupvisorsInstanceId('', supvisors)
    assert sup_id.identifier is None
    assert sup_id.nick_identifier is None
    assert sup_id.host_id is None
    assert sup_id.http_port == 25000  # defaulted to supervisor server port
    assert sup_id.event_port == 25001  # defaulted to http_port + 1
    assert sup_id.host_name is None
    assert sup_id.aliases is None
    assert sup_id.ip_addresses is None
    assert sup_id.ip_address is None
    assert sup_id.stereotypes == []
    with pytest.raises(TypeError):
        str(sup_id)
    assert sup_id.serial() == {'host_id': None, 'host_name': None, 'http_port': 25000,
                               'identifier': None, 'ip_addresses': None,
                               'nick_identifier': None, 'stereotypes': []}


def test_sup_id_create_host(supvisors):
    """ Test the values set at SupvisorsInstanceId construction. """
    # test with simple host name match
    sup_id = SupvisorsInstanceId('10.0.0.1', supvisors)
    assert sup_id.identifier == '10.0.0.1:25000'
    assert sup_id.nick_identifier == '10.0.0.1'
    assert sup_id.host_id == '10.0.0.1'
    assert sup_id.http_port == 25000  # defaulted to supervisor server port
    assert sup_id.event_port == 25200  # defaulted to options.event_port
    assert sup_id.host_name == '10.0.0.1'
    assert sup_id.aliases == ['10.0.0.1']
    assert sup_id.ip_addresses == ['10.0.0.1']
    assert sup_id.ip_address == '10.0.0.1'
    assert sup_id.stereotypes == []
    assert str(sup_id) == '<10.0.0.1>10.0.0.1:25000'
    assert sup_id.serial() == {'host_id': '10.0.0.1', 'host_name': '10.0.0.1', 'http_port': 25000,
                               'identifier': '10.0.0.1:25000', 'ip_addresses': ['10.0.0.1'],
                               'nick_identifier': '10.0.0.1', 'stereotypes': []}


def test_sup_id_create_host_port(supvisors):
    """ Test the values set at SupvisorsInstanceId construction. """
    # test with host+ports match
    sup_id = SupvisorsInstanceId('10.0.0.1:7777', supvisors)
    assert sup_id.identifier == '10.0.0.1:7777'
    assert sup_id.nick_identifier == '10.0.0.1:7777'
    assert sup_id.host_id == '10.0.0.1'
    assert sup_id.http_port == 7777
    assert sup_id.event_port == 25200  # defaulted to options.event_port
    assert sup_id.host_name == '10.0.0.1'
    assert sup_id.aliases == ['10.0.0.1']
    assert sup_id.ip_addresses == ['10.0.0.1']
    assert sup_id.ip_address == '10.0.0.1'
    assert sup_id.stereotypes == []
    assert str(sup_id) == '10.0.0.1:7777'
    assert sup_id.serial() == {'host_id': '10.0.0.1', 'host_name': '10.0.0.1', 'http_port': 7777,
                               'identifier': '10.0.0.1:7777', 'ip_addresses': ['10.0.0.1'],
                               'nick_identifier': '10.0.0.1:7777', 'stereotypes': []}
    # test with host+ports match (internal port defined)
    supvisors.options.event_port = 0
    sup_id = SupvisorsInstanceId('10.0.0.1:7777', supvisors)
    assert sup_id.identifier == '10.0.0.1:7777'
    assert sup_id.nick_identifier == '10.0.0.1:7777'
    assert sup_id.host_id == '10.0.0.1'
    assert sup_id.http_port == 7777
    assert sup_id.event_port == 7778  # defaulted to http_port + 1
    assert sup_id.host_name == '10.0.0.1'
    assert sup_id.aliases == ['10.0.0.1']
    assert sup_id.ip_addresses == ['10.0.0.1']
    assert sup_id.ip_address == '10.0.0.1'
    assert sup_id.stereotypes == []
    assert str(sup_id) == '10.0.0.1:7777'
    assert sup_id.serial() == {'host_id': '10.0.0.1', 'host_name': '10.0.0.1', 'http_port': 7777,
                               'identifier': '10.0.0.1:7777', 'ip_addresses': ['10.0.0.1'],
                               'nick_identifier': '10.0.0.1:7777', 'stereotypes': []}


def test_sup_id_create_identifier(supvisors):
    """ Test the values set at SupvisorsInstanceId construction. """
    # test with identifier set and only host name
    sup_id = SupvisorsInstanceId('<supvisors>cliche81', supvisors)
    assert sup_id.identifier == 'cliche81:25000'
    assert sup_id.nick_identifier == 'supvisors'
    assert sup_id.host_id == 'cliche81'
    assert sup_id.http_port == 25000
    assert sup_id.event_port == 25200  # defaulted to options.event_port
    assert sup_id.host_name == 'cliche81'
    assert sup_id.aliases == ['cliche81']
    assert sup_id.ip_addresses == ['cliche81']
    assert sup_id.ip_address == 'cliche81'
    assert sup_id.stereotypes == []
    assert str(sup_id) == '<supvisors>cliche81:25000'
    assert sup_id.serial() == {'host_id': 'cliche81', 'host_name': 'cliche81', 'http_port': 25000,
                               'identifier': 'cliche81:25000', 'ip_addresses': ['cliche81'],
                               'nick_identifier': 'supvisors', 'stereotypes': []}
    # test with identifier, host name and http port set
    sup_id = SupvisorsInstanceId('<supvisors>cliche81:8888', supvisors)
    assert sup_id.identifier == 'cliche81:8888'
    assert sup_id.nick_identifier == 'supvisors'
    assert sup_id.host_id == 'cliche81'
    assert sup_id.http_port == 8888
    assert sup_id.event_port == 25200  # defaulted to options.event_port
    assert sup_id.host_name == 'cliche81'
    assert sup_id.aliases == ['cliche81']
    assert sup_id.ip_addresses == ['cliche81']
    assert sup_id.ip_address == 'cliche81'
    assert sup_id.stereotypes == []
    assert str(sup_id) == '<supvisors>cliche81:8888'
    assert sup_id.serial() == {'host_id': 'cliche81', 'host_name': 'cliche81', 'http_port': 8888,
                               'identifier': 'cliche81:8888', 'ip_addresses': ['cliche81'],
                               'nick_identifier': 'supvisors', 'stereotypes': []}


def test_sup_id_host_matches(supvisors):
    """ Test the SupvisorsInstanceId.host_matches method. """
    sup_id = SupvisorsInstanceId('<supvisors>cliche81', supvisors)
    sup_id.aliases = ['cliche', 'rocky51']
    assert sup_id.host_matches('cliche81')
    assert sup_id.host_matches('cliche')
    assert sup_id.host_matches('rocky51')
    assert not sup_id.host_matches('rocky')


@pytest.fixture
def mapper(supvisors):
    """ Return the instance to test. """
    return SupvisorsMapper(supvisors)


def test_mapper_create(supvisors, mapper):
    """ Test the values set at construction. """
    assert mapper.supvisors is supvisors
    assert mapper.logger is supvisors.logger
    assert mapper.instances == {}
    assert mapper.nodes == {}
    assert mapper.local_identifier is None
    assert mapper.initial_identifiers == []


def test_get_nick_identifier(supvisors):
    """ Test the SupvisorsMapper.get_nick_identifier method. """
    mapper = supvisors.mapper
    assert mapper.get_nick_identifier(mapper.local_identifier) == mapper.local_nick_identifier
    assert mapper.get_nick_identifier('10.0.0.1:25000') == '10.0.0.1'
    assert mapper.get_nick_identifier('10.0.0.2:25000') == '10.0.0.2'
    assert mapper.get_nick_identifier('10.0.0.3:25000') == '10.0.0.3'
    assert mapper.get_nick_identifier('10.0.0.4:25000') == '10.0.0.4'
    assert mapper.get_nick_identifier('10.0.0.5:25000') == '10.0.0.5'
    assert mapper.get_nick_identifier(f'{socket.getfqdn()}:15000') == 'test'


def test_add_instance(mapper):
    """ Test the SupvisorsMapper.add_instance method. """
    # test error (no IP address)
    with pytest.raises(ValueError):
        mapper.add_instance('<dummy>:1234:')
    assert mapper.instances == {}
    assert mapper._nick_identifiers == {}
    assert mapper.nodes == {}
    with pytest.raises(ValueError):
        mapper.add_instance('<dummy>:1234:4321')
    assert mapper.instances == {}
    assert mapper._nick_identifiers == {}
    assert mapper.nodes == {}
    # test correct format
    mapper.add_instance('<dummy_2>10.0.0.1:1234')
    assert sorted(mapper.instances.keys()) == ['10.0.0.1:1234']
    assert mapper._nick_identifiers == {'dummy_2': '10.0.0.1:1234'}
    assert mapper.nodes == {'10.0.0.1': ['10.0.0.1:1234']}
    mapper.add_instance('10.0.0.1:4321')
    assert sorted(mapper.instances.keys()) == ['10.0.0.1:1234', '10.0.0.1:4321']
    assert mapper._nick_identifiers == {'dummy_2': '10.0.0.1:1234', '10.0.0.1:4321': '10.0.0.1:4321'}
    assert mapper.nodes == {'10.0.0.1': ['10.0.0.1:1234', '10.0.0.1:4321']}
    mapper.add_instance('10.0.0.2')
    assert sorted(mapper.instances.keys()) == ['10.0.0.1:1234', '10.0.0.1:4321', '10.0.0.2:25000']
    assert mapper._nick_identifiers == {'dummy_2': '10.0.0.1:1234', '10.0.0.1:4321': '10.0.0.1:4321',
                                        '10.0.0.2': '10.0.0.2:25000'}
    assert mapper.nodes == {'10.0.0.1': ['10.0.0.1:1234', '10.0.0.1:4321'], '10.0.0.2': ['10.0.0.2:25000']}
    mapper.add_instance('<dummy_1>10.0.0.3')
    assert sorted(mapper.instances.keys()) == ['10.0.0.1:1234', '10.0.0.1:4321', '10.0.0.2:25000', '10.0.0.3:25000']
    assert mapper._nick_identifiers == {'dummy_2': '10.0.0.1:1234', '10.0.0.1:4321': '10.0.0.1:4321',
                                        '10.0.0.2': '10.0.0.2:25000', 'dummy_1': '10.0.0.3:25000'}
    assert mapper.nodes == {'10.0.0.1': ['10.0.0.1:1234', '10.0.0.1:4321'], '10.0.0.2': ['10.0.0.2:25000'],
                            '10.0.0.3': ['10.0.0.3:25000']}


def test_mapper_configure(mocker, mapper):
    """ Test the storage of the expected Supvisors instances. """
    mocked_find = mocker.patch.object(mapper, 'find_local_identifier')
    # configure mapper with elements
    items = ['127.0.0.1', '<supervisor_05>10.0.0.5:7777', '10.0.0.4:15000', '10.0.0.5:9999']
    core_items = ['127.0.0.1', 'supervisor_05', '10.0.0.4:15000', 'supervisor_06', '10.0.0.4']
    mapper.configure(items, 'test', core_items)
    assert list(mapper.instances.keys()) == ['127.0.0.1:25000', '10.0.0.5:7777', '10.0.0.4:15000', '10.0.0.5:9999']
    assert mapper._nick_identifiers == {'10.0.0.4:15000': '10.0.0.4:15000',
                                        '10.0.0.5:9999': '10.0.0.5:9999',
                                        '127.0.0.1': '127.0.0.1:25000',
                                        'supervisor_05': '10.0.0.5:7777'}
    assert mapper._core_identifiers == core_items
    assert mapper.core_identifiers == ['127.0.0.1:25000', '10.0.0.5:7777', '10.0.0.4:15000']
    assert mapper.nodes == {'10.0.0.4': ['10.0.0.4:15000'],
                            '10.0.0.5': ['10.0.0.5:7777', '10.0.0.5:9999'],
                            '127.0.0.1': ['127.0.0.1:25000']}
    assert mapper.initial_identifiers == ['127.0.0.1:25000', '10.0.0.5:7777', '10.0.0.4:15000', '10.0.0.5:9999']
    assert mocked_find.call_args_list == [call('test')]
    mocked_find.reset_mock()
    # configure mapper with one invalid element
    items = ['127.0.0.1', '<dummy_id>', '<supervisor_05>10.0.0.5', '10.0.0.4:15000']
    with pytest.raises(ValueError):
        mapper.configure(items, '', core_items)
    assert not mocked_find.called


def test_find_local_identifier_fqdn(supvisors, mapper):
    """ Test the SupvisorsMapper.find_local_identifier method when the item host name matches the . """
    fqdn = getfqdn()
    items = [fqdn, '<supervisor>10.0.0.5:7777']
    for item in items:
        supvisors_id = SupvisorsInstanceId(item, supvisors)
        mapper._instances[supvisors_id.identifier] = supvisors_id
        assert supvisors_id.stereotypes == []
    # 1. force host name and aliases
    sup_id = mapper._instances[f'{fqdn}:25000']
    sup_id.host_name = fqdn
    sup_id.aliases = ['dummy']
    # find self
    mapper.find_local_identifier({'test'})
    assert mapper.local_identifier == f'{fqdn}:25000'
    for identifier, supvisors_id in mapper.instances.items():
        assert identifier != fqdn or supvisors_id.stereotypes == ['test']
    # 2. move fqdn
    sup_id.host_name = 'dummy'
    sup_id.aliases = [fqdn]
    sup_id.stereotypes = []
    # find self
    mapper.find_local_identifier({'test_2'})
    assert mapper.local_identifier == f'{fqdn}:25000'
    for identifier, supvisors_id in mapper.instances.items():
        assert identifier != fqdn or supvisors_id.stereotypes == ['test_2']


def test_find_local_identifier_host_name(mapper):
    """ Test the SupvisorsMapper.find_local_identifier method when one instance matches the host name and the HTTP
    server port. """
    host_name, fqdn = gethostname(), getfqdn()
    items = ['127.0.0.1', f'{host_name}:25000']
    for item in items:
        supvisors_id = SupvisorsInstanceId(item, mapper.supvisors)
        mapper._instances[supvisors_id.identifier] = supvisors_id
        assert supvisors_id.stereotypes == []
    # 1. force host name and aliases
    sup_id = mapper._instances[f'{host_name}:25000']
    sup_id.host_name = fqdn
    sup_id.aliases = ['dummy']
    # find self
    mapper.find_local_identifier({'test'})
    assert mapper.local_identifier == f'{host_name}:25000'
    for identifier, supvisors_id in mapper.instances.items():
        assert identifier != fqdn or supvisors_id.stereotype == ['test']
    # 2. move fqdn
    sup_id.host_name = 'dummy'
    sup_id.aliases = [fqdn]
    # find self
    mapper.find_local_identifier({'test_2'})
    assert mapper.local_identifier == f'{host_name}:25000'
    for identifier, supvisors_id in mapper.instances.items():
        assert identifier != fqdn or supvisors_id.stereotype == ['test_2']


def test_find_local_identifier_ip_address(supvisors, mapper):
    """ Test the SupvisorsMapper.find_local_identifier method when one instance matches the IP address of the host
    and the HTTP server port. """
    host_name, fqdn = gethostname(), getfqdn()
    _, _, ip_addresses = get_addresses(fqdn, supvisors.logger)
    items = ['127.0.0.1', f'<host>{ip_addresses[0]}:25000']
    for item in items:
        supvisors_id = SupvisorsInstanceId(item, supvisors)
        mapper._instances[supvisors_id.identifier] = supvisors_id
        assert supvisors_id.stereotypes == []
    # 1. force host name and aliases
    sup_id = mapper._instances[f'{ip_addresses[0]}:25000']
    sup_id.host_name = fqdn
    sup_id.aliases = ['dummy']
    # find self
    mapper.find_local_identifier({'test'})
    assert mapper.local_identifier == f'{ip_addresses[0]}:25000'
    for identifier, supvisors_id in mapper.instances.items():
        assert identifier != fqdn or supvisors_id.stereotypes == ['test']
    # 2. move fqdn
    sup_id.host_name = 'dummy'
    sup_id.aliases = [fqdn]
    # find self
    mapper.find_local_identifier({'test_2'})
    assert mapper.local_identifier == f'{ip_addresses[0]}:25000'
    for identifier, supvisors_id in mapper.instances.items():
        assert identifier != fqdn or supvisors_id.stereotypes == ['test_2']


def test_find_local_identifier_inconsistent(supvisors, mapper):
    """ Test the SupvisorsMapper.find_local_identifier method when the identifier corresponds to thr local Supvisors
    instance but the host_name and http_port do not match. """
    host_name = gethostname()
    supvisors_id = SupvisorsInstanceId(f'<{host_name}>10.0.0.5:7777', supvisors)
    mapper._instances[supvisors_id.identifier] = supvisors_id
    assert supvisors_id.stereotypes == []
    # find self
    print(supvisors.supervisor_data.identifier)
    with pytest.raises(ValueError):
        mapper.find_local_identifier({'test'})
    assert supvisors_id.stereotypes == []


def test_find_local_identifier_multiple(supvisors, mapper):
    """ Test the SupvisorsMapper.find_local_identifier method when more than one instance matches the host name
    or IP address of the host. """
    host_name = gethostname()
    fqdn = getfqdn()
    items = ['10.0.0.1', f'{host_name}:25000']
    for item in items:
        supvisors_id = SupvisorsInstanceId(item, supvisors)
        mapper._instances[supvisors_id.identifier] = supvisors_id
        assert supvisors_id.stereotypes == []
        # update default processing of host
        supvisors_id.host_name = fqdn
    # find self
    with pytest.raises(ValueError):
        mapper.find_local_identifier({'test'})
    for supvisors_id in mapper._instances.values():
        assert supvisors_id.stereotypes == []


def test_find_local_identifier_none(supvisors, mapper):
    """ Test the SupvisorsMapper.find_local_identifier method when no instance matches the host name
    or IP address of the host. """
    host_name = gethostname()
    items = ['10.0.0.2', '<dummy>cliche:60000', f'{host_name}:60000']
    for item in items:
        supvisors_id = SupvisorsInstanceId(item, supvisors)
        mapper._instances[supvisors_id.identifier] = supvisors_id
    with pytest.raises(ValueError):
        mapper.find_local_identifier({'test'})
    for supvisors_id in mapper._instances.values():
        assert supvisors_id.stereotypes == []


def test_assign_stereotypes(supvisors, mapper):
    """ Test the SupvisorsMapper.assign_stereotypes method. """
    # add context
    items = ['10.0.0.1', '10.0.0.3', '10.0.0.4', '10.0.0.2']
    for item in items:
        supvisors_id = SupvisorsInstanceId(item, supvisors)
        mapper._instances[supvisors_id.identifier] = supvisors_id
        assert supvisors_id.stereotypes == []
    assert mapper.stereotypes == {}
    # test with empty stereotype
    mapper.assign_stereotypes('10.0.0.1:25000', set())
    assert mapper.stereotypes == {}
    assert all(x.stereotypes == [] for x in mapper.instances.values())
    # assign stereotype to a Supvisors instance
    mapper.assign_stereotypes('10.0.0.3:25000', {'test'})
    assert mapper.stereotypes == {'test': ['10.0.0.3:25000']}
    assert mapper.instances['10.0.0.1:25000'].stereotypes == []
    assert mapper.instances['10.0.0.2:25000'].stereotypes == []
    assert mapper.instances['10.0.0.3:25000'].stereotypes == ['test']
    assert mapper.instances['10.0.0.4:25000'].stereotypes == []
    # assign same stereotype to another Supvisors instance that is before in the instances order
    mapper.assign_stereotypes('10.0.0.1:25000', {'test'})
    assert mapper.stereotypes == {'test': ['10.0.0.1:25000', '10.0.0.3:25000']}
    assert mapper.instances['10.0.0.1:25000'].stereotypes == ['test']
    assert mapper.instances['10.0.0.2:25000'].stereotypes == []
    assert mapper.instances['10.0.0.3:25000'].stereotypes == ['test']
    assert mapper.instances['10.0.0.4:25000'].stereotypes == []
    # assign same stereotype to another Supvisors instance that is after in the instances order
    mapper.assign_stereotypes('10.0.0.2:25000', {'test', 'other test'})
    assert mapper.stereotypes == {'test': ['10.0.0.1:25000', '10.0.0.3:25000', '10.0.0.2:25000'],
                                  'other test': ['10.0.0.2:25000']}
    assert mapper.instances['10.0.0.1:25000'].stereotypes == ['test']
    assert sorted(mapper.instances['10.0.0.2:25000'].stereotypes) == sorted(['test', 'other test'])
    assert mapper.instances['10.0.0.3:25000'].stereotypes == ['test']
    assert mapper.instances['10.0.0.4:25000'].stereotypes == []
    # assign another stereotype to another Supvisors instance
    mapper.assign_stereotypes('10.0.0.4:25000', {'other test'})
    assert mapper.stereotypes == {'test': ['10.0.0.1:25000', '10.0.0.3:25000', '10.0.0.2:25000'],
                                  'other test': ['10.0.0.4:25000', '10.0.0.2:25000']}
    assert mapper.instances['10.0.0.1:25000'].stereotypes == ['test']
    assert sorted(mapper.instances['10.0.0.2:25000'].stereotypes) == sorted(['test', 'other test'])
    assert mapper.instances['10.0.0.3:25000'].stereotypes == ['test']
    assert mapper.instances['10.0.0.4:25000'].stereotypes == ['other test']
    # test serial with stereotypes
    assert mapper.instances['10.0.0.1:25000'].serial()['stereotypes'] == ['test']
    assert sorted(mapper.instances['10.0.0.2:25000'].serial()['stereotypes']) == ['other test', 'test']
    assert mapper.instances['10.0.0.3:25000'].serial()['stereotypes'] == ['test']
    assert mapper.instances['10.0.0.4:25000'].serial()['stereotypes'] == ['other test']


def test_filter(supvisors, mapper):
    """ Test the SupvisorsMapper.filter method. """
    # add context
    hostname = gethostname()
    items = ['127.0.0.1', '<host>10.0.0.1:2222', f'{hostname}:60000', '10.0.0.2']
    for item in items:
        supvisors_id = SupvisorsInstanceId(item, supvisors)
        mapper._instances[supvisors_id.identifier] = supvisors_id
        mapper._nick_identifiers[supvisors_id.nick_identifier] = supvisors_id.identifier
    # force host name to FQDN
    mapper._instances[f'{hostname}:60000'].host_name = getfqdn()
    # test with a bunch of identifiers
    identifier_list = ['host', f'{hostname}:60000', hostname, 'host', 'supervisor', '10.0.0.1']
    assert mapper.filter(identifier_list) == ['10.0.0.1:2222', f'{hostname}:60000']
    # set stereotype to one of them
    mapper.assign_stereotypes('10.0.0.2:25000', {'supervisor'})
    assert mapper.filter(identifier_list) == ['10.0.0.1:2222', f'{hostname}:60000', '10.0.0.2:25000']


def test_check_candidate(supvisors, mapper):
    """ Test the SupvisorsMapper.check_candidate method. """
    # add context
    hostname = gethostname()
    items = ['127.0.0.1', '<host>10.0.0.1:2222', f'{hostname}:60000', '10.0.0.2']
    for item in items:
        supvisors_id = SupvisorsInstanceId(item, supvisors)
        mapper._instances[supvisors_id.identifier] = supvisors_id
        mapper._nick_identifiers[supvisors_id.nick_identifier] = supvisors_id.identifier
    # force host name to FQDN
    mapper._instances[f'{hostname}:60000'].host_name = getfqdn()
    # test existing instances
    assert not mapper.check_candidate('127.0.0.1:25000', '127.0.0.1')
    assert not mapper.check_candidate('10.0.0.1:2222', 'host')
    assert not mapper.check_candidate(f'{hostname}:60000', f'{hostname}:60000')
    assert not mapper.check_candidate('10.0.0.2:25000', '10.0.0.2')
    # test partially known instances (this will trigger log warnings)
    assert not mapper.check_candidate('127.0.0.1:25000', 'host')
    assert not mapper.check_candidate(f'{hostname}:50000', '10.0.0.2')
    # test unknown instances
    assert mapper.check_candidate(f'{hostname}:50000', f'{hostname}:50000')
    assert mapper.check_candidate('10.0.0.3:3333', '10.0.0.3')
